package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.DocumentIndex
import solutions.bitbadger.documents.query.DefinitionQuery
import java.sql.Connection
import kotlin.jvm.Throws

/**
 * Functions to define tables and indexes
 */
object Definition {

    /**
     * Create a document table if necessary
     *
     * @param tableName The table whose existence should be ensured (may include schema)
     * @param conn The connection on which the query should be executed
     * @throws DocumentException If the dialect is not configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun ensureTable(tableName: String, conn: Connection) =
        Configuration.dialect("ensure $tableName exists").let {
            Custom.nonQuery(DefinitionQuery.ensureTable(tableName, it), conn = conn)
            Custom.nonQuery(DefinitionQuery.ensureKey(tableName, it), conn = conn)
        }

    /**
     * Create a document table if necessary
     *
     * @param tableName The table whose existence should be ensured (may include schema)
     * @throws DocumentException If no connection string has been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun ensureTable(tableName: String) =
        Configuration.dbConn().use { ensureTable(tableName, it) }

    /**
     * Create an index on field(s) within documents in the specified table if necessary
     *
     * @param tableName The table to be indexed (may include schema)
     * @param indexName The name of the index to create
     * @param fields One or more fields to be indexed
     * @param conn The connection on which the query should be executed
     * @throws DocumentException If any dependent process does
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun ensureFieldIndex(tableName: String, indexName: String, fields: Collection<String>, conn: Connection) =
        Custom.nonQuery(DefinitionQuery.ensureIndexOn(tableName, indexName, fields), conn = conn)

    /**
     * Create an index on field(s) within documents in the specified table if necessary
     *
     * @param tableName The table to be indexed (may include schema)
     * @param indexName The name of the index to create
     * @param fields One or more fields to be indexed
     * @throws DocumentException If no connection string has been set, or if any dependent process does
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun ensureFieldIndex(tableName: String, indexName: String, fields: Collection<String>) =
        Configuration.dbConn().use { ensureFieldIndex(tableName, indexName, fields, it) }

    /**
     * Create a document index on a table (PostgreSQL only)
     *
     * @param tableName The table to be indexed (may include schema)
     * @param indexType The type of index to ensure
     * @param conn The connection on which the query should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun ensureDocumentIndex(tableName: String, indexType: DocumentIndex, conn: Connection) =
        Custom.nonQuery(DefinitionQuery.ensureDocumentIndexOn(tableName, indexType), conn = conn)

    /**
     * Create a document index on a table (PostgreSQL only)
     *
     * @param tableName The table to be indexed (may include schema)
     * @param indexType The type of index to ensure
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun ensureDocumentIndex(tableName: String, indexType: DocumentIndex) =
        Configuration.dbConn().use { ensureDocumentIndex(tableName, indexType, it) }
}
