package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.query.FindQuery
import solutions.bitbadger.documents.query.orderBy
import java.sql.Connection
import java.util.Optional
import kotlin.jvm.Throws

/**
 * Functions to find and retrieve documents
 */
object Find {

    /**
     * Retrieve all documents in the given table, ordering results by the optional given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents from the given table
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> all(tableName: String, clazz: Class<TDoc>, orderBy: Collection<Field<*>>? = null, conn: Connection) =
        Custom.list(FindQuery.all(tableName) + (orderBy?.let(::orderBy) ?: ""), listOf(), clazz, conn, Results::fromData)

    /**
     * Retrieve all documents in the given table
     *
     * @param tableName The table from which documents should be retrieved
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents from the given table
     * @throws DocumentException If no connection string has been set, or if query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> all(tableName: String, clazz: Class<TDoc>, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { all(tableName, clazz, orderBy, it) }

    /**
     * Retrieve all documents in the given table
     *
     * @param tableName The table from which documents should be retrieved
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents from the given table
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> all(tableName: String, clazz: Class<TDoc>, conn: Connection) =
        all(tableName, clazz, null, conn)

    /**
     * Retrieve a document by its ID
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId The ID of the document to retrieve
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item with the document if it is found
     * @throws DocumentException If no dialect has been configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey, TDoc> byId(tableName: String, docId: TKey, clazz: Class<TDoc>, conn: Connection) =
        Custom.single(
            FindQuery.byId(tableName, docId),
            Parameters.addFields(listOf(Field.equal(Configuration.idField, docId, ":id"))),
            clazz,
            conn,
            Results::fromData
        )

    /**
     * Retrieve a document by its ID
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId The ID of the document to retrieve
     * @param clazz The class of the document to be returned
     * @return An `Optional` item with the document if it is found
     * @throws DocumentException If no connection string has been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey, TDoc> byId(tableName: String, docId: TKey, clazz: Class<TDoc>) =
        Configuration.dbConn().use { byId(tableName, docId, clazz, it) }

    /**
     * Retrieve documents using a field comparison, ordering results by the given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the field comparison
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ): List<TDoc> {
        val named = Parameters.nameFields(fields)
        return Custom.list(
            FindQuery.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            clazz,
            conn,
            Results::fromData
        )
    }

    /**
     * Retrieve documents using a field comparison, ordering results by the given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the field comparison
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { byFields(tableName, fields, clazz, howMatched, orderBy, it) }

    /**
     * Retrieve documents using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the field comparison
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) =
        byFields(tableName, fields, clazz, howMatched, null, conn)

    /**
     * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> byContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        Custom.list(
            FindQuery.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameters.json(":criteria", criteria)),
            clazz,
            conn,
            Results::fromData
        )

    /**
     * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc, TContains> byContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { byContains(tableName, criteria, clazz, orderBy, it) }

    /**
     * Retrieve documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> byContains(tableName: String, criteria: TContains, clazz: Class<TDoc>, conn: Connection) =
        byContains(tableName, criteria, clazz, null, conn)

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> byJsonPath(
        tableName: String,
        path: String,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        Custom.list(
            FindQuery.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameter(":path", ParameterType.STRING, path)),
            clazz,
            conn,
            Results::fromData
        )

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> byJsonPath(tableName: String, path: String, clazz: Class<TDoc>, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { byJsonPath(tableName, path, clazz, orderBy, it) }

    /**
     * Retrieve documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> byJsonPath(tableName: String, path: String, clazz: Class<TDoc>, conn: Connection) =
        byJsonPath(tableName, path, clazz, null, conn)

    /**
     * Retrieve the first document using a field comparison and optional ordering fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item, with the first document matching the field comparison if found
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ): Optional<TDoc & Any> {
        val named = Parameters.nameFields(fields)
        return Custom.single(
            FindQuery.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            clazz,
            conn,
            Results::fromData
        )
    }

    /**
     * Retrieve the first document using a field comparison and optional ordering fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return An `Optional` item, with the first document matching the field comparison if found
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { firstByFields(tableName, fields, clazz, howMatched, orderBy, it) }

    /**
     * Retrieve the first document using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param clazz The class of the document to be returned
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item, with the first document matching the field comparison if found
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        clazz: Class<TDoc>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) =
        firstByFields(tableName, fields, clazz, howMatched, null, conn)

    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item, with the first document matching the JSON containment query if found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        Custom.single(
            FindQuery.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameters.json(":criteria", criteria)),
            clazz,
            conn,
            Results::fromData
        )

    /**
     * Retrieve the first document using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item, with the first document matching the JSON containment query if found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc, TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        conn: Connection
    ) =
        firstByContains(tableName, criteria, clazz, null, conn)

    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return An `Optional` item, with the first document matching the JSON containment query if found
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc, TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { firstByContains(tableName, criteria, clazz, orderBy, it) }

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item, with the first document matching the JSON Path match query if found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> firstByJsonPath(
        tableName: String,
        path: String,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) =
        Custom.single(
            FindQuery.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameter(":path", ParameterType.STRING, path)),
            clazz,
            conn,
            Results::fromData
        )

    /**
     * Retrieve the first document using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param conn The connection over which documents should be retrieved
     * @return An `Optional` item, with the first document matching the JSON Path match query if found
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TDoc> firstByJsonPath(tableName: String, path: String, clazz: Class<TDoc>, conn: Connection) =
        firstByJsonPath(tableName, path, clazz, null, conn)

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param clazz The class of the document to be returned
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return An `Optional` item, with the first document matching the JSON Path match query if found
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TDoc> firstByJsonPath(
        tableName: String,
        path: String,
        clazz: Class<TDoc>,
        orderBy: Collection<Field<*>>? = null
    ) =
        Configuration.dbConn().use { firstByJsonPath(tableName, path, clazz, orderBy, it) }
}
