package solutions.bitbadger.documents.java

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.query.FindQuery
import solutions.bitbadger.documents.query.orderBy
import java.io.PrintWriter
import java.sql.Connection
import kotlin.jvm.Throws

/**
 * Functions to find and retrieve documents, returning them as JSON strings
 */
object Json {

    /**
     * Retrieve all documents in the given table, ordering results by the optional given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents from the given table
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun all(tableName: String, orderBy: Collection<Field<*>>? = null, conn: Connection) =
        Custom.jsonArray(
            FindQuery.all(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(),
            conn,
            Results::jsonFromData
        )

    /**
     * Retrieve all documents in the given table (creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A JSON array of documents from the given table
     * @throws DocumentException If no connection string has been set, or if query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun all(tableName: String, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { all(tableName, orderBy, it) }

    /**
     * Retrieve all documents in the given table
     *
     * @param tableName The table from which documents should be retrieved
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents from the given table
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun all(tableName: String, conn: Connection) =
        all(tableName, null, conn)

    /**
     * Write all documents in the given table to the given `PrintWriter`, ordering results by the optional given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeAll(tableName: String, writer: PrintWriter, orderBy: Collection<Field<*>>? = null, conn: Connection) =
        Custom.writeJsonArray(
            FindQuery.all(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(),
            writer,
            conn,
            Results::jsonFromData
        )

    /**
     * Write all documents in the given table to the given `PrintWriter`, ordering results by the optional given fields
     * (creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun writeAll(tableName: String, writer: PrintWriter, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { writeAll(tableName, writer, orderBy, it) }

    /**
     * Write all documents in the given table to the given `PrintWriter`
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If query execution fails
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeAll(tableName: String, writer: PrintWriter, conn: Connection) =
        writeAll(tableName, writer, null, conn)

    /**
     * Retrieve a document by its ID
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId The ID of the document to retrieve
     * @param conn The connection over which documents should be retrieved
     * @return A JSON document if found, an empty JSON object if not found
     * @throws DocumentException If no dialect has been configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey> byId(tableName: String, docId: TKey, conn: Connection) =
        Custom.jsonSingle(
            FindQuery.byId(tableName, docId),
            Parameters.addFields(listOf(Field.equal(Configuration.idField, docId, ":id"))),
            conn,
            Results::jsonFromData
        )

    /**
     * Retrieve a document by its ID (creates connection)
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId The ID of the document to retrieve
     * @return A JSON document if found, an empty JSON object if not found
     * @throws DocumentException If no connection string has been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey> byId(tableName: String, docId: TKey) =
        Configuration.dbConn().use { byId(tableName, docId, it) }

    /**
     * Write a document to the given `PrintWriter` by its ID (writes empty object if not found)
     *
     * @param tableName The table from which the document should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param docId The ID of the document to retrieve
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If no dialect has been configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey> writeById(tableName: String, writer: PrintWriter, docId: TKey, conn: Connection) =
        writer.write(byId(tableName, docId, conn))

    /**
     * Write a document to the given `PrintWriter` by its ID (writes empty object if not found; creates connection)
     *
     * @param tableName The table from which the document should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param docId The ID of the document to retrieve
     * @throws DocumentException If no dialect has been configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TKey> writeById(tableName: String, writer: PrintWriter, docId: TKey) =
        Configuration.dbConn().use { writeById(tableName, writer, docId, it) }

    /**
     * Retrieve documents using a field comparison, ordering results by the given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents matching the field comparison
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ): String {
        val named = Parameters.nameFields(fields)
        return Custom.jsonArray(
            FindQuery.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            conn,
            Results::jsonFromData
        )
    }

    /**
     * Retrieve documents using a field comparison, ordering results by the given fields (creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A JSON array of documents matching the field comparison
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { byFields(tableName, fields, howMatched, orderBy, it) }

    /**
     * Retrieve documents using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents matching the field comparison
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null, conn: Connection) =
        byFields(tableName, fields, howMatched, null, conn)

    /**
     * Write documents to the given `PrintWriter` using a field comparison, ordering results by the given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeByFields(
        tableName: String,
        writer: PrintWriter,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) {
        val named = Parameters.nameFields(fields)
        Custom.writeJsonArray(
            FindQuery.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            writer,
            conn,
            Results::jsonFromData
        )
    }

    /**
     * Write documents to the given `PrintWriter` using a field comparison, ordering results by the given fields
     * (creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun writeByFields(
        tableName: String,
        writer: PrintWriter,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { writeByFields(tableName, writer, fields, howMatched, orderBy, it) }

    /**
     * Write documents to the given `PrintWriter` using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeByFields(
        tableName: String,
        writer: PrintWriter,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) = writeByFields(tableName, writer, fields, howMatched, null, conn)

    /**
     * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> byContains(
        tableName: String,
        criteria: TContains,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = Custom.jsonArray(
        FindQuery.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
        listOf(Parameters.json(":criteria", criteria)),
        conn,
        Results::jsonFromData
    )

    /**
     * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only; creates
     * connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A JSON array of documents matching the JSON containment query
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TContains> byContains(tableName: String, criteria: TContains, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { byContains(tableName, criteria, orderBy, it) }

    /**
     * Retrieve documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> byContains(tableName: String, criteria: TContains, conn: Connection) =
        byContains(tableName, criteria, null, conn)

    /**
     * Write documents to the given `PrintWriter` using a JSON containment query, ordering results by the given fields
     * (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> writeByContains(
        tableName: String,
        writer: PrintWriter,
        criteria: TContains,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = Custom.writeJsonArray(
        FindQuery.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
        listOf(Parameters.json(":criteria", criteria)),
        writer,
        conn,
        Results::jsonFromData
    )

    /**
     * Write documents to the given `PrintWriter` using a JSON containment query, ordering results by the given fields
     * (PostgreSQL only; creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TContains> writeByContains(
        tableName: String,
        writer: PrintWriter,
        criteria: TContains,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { writeByContains(tableName, writer, criteria, orderBy, it) }

    /**
     * Write documents to the given `PrintWriter` using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> writeByContains(tableName: String, writer: PrintWriter, criteria: TContains, conn: Connection) =
        writeByContains(tableName, writer, criteria, null, conn)

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String, path: String, orderBy: Collection<Field<*>>? = null, conn: Connection) =
        Custom.jsonArray(
            FindQuery.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameter(":path", ParameterType.STRING, path)),
            conn,
            Results::jsonFromData
        )

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only; creates
     * connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A JSON array of documents matching the JSON Path match query
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byJsonPath(tableName: String, path: String, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { byJsonPath(tableName, path, orderBy, it) }

    /**
     * Retrieve documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param conn The connection over which documents should be retrieved
     * @return A JSON array of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String, path: String, conn: Connection) =
        byJsonPath(tableName, path, null, conn)

    /**
     * Write documents to the given `PrintWriter` using a JSON Path match query, ordering results by the given fields
     * (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeByJsonPath(
        tableName: String,
        writer: PrintWriter,
        path: String,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = Custom.writeJsonArray(
        FindQuery.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
        listOf(Parameter(":path", ParameterType.STRING, path)),
        writer,
        conn,
        Results::jsonFromData
    )

    /**
     * Write documents to the given `PrintWriter` using a JSON Path match query, ordering results by the given fields
     * (PostgreSQL only; creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun writeByJsonPath(tableName: String, writer: PrintWriter, path: String, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { writeByJsonPath(tableName, writer, path, orderBy, it) }

    /**
     * Write documents to the given `PrintWriter` using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param path The JSON path comparison to match
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeByJsonPath(tableName: String, writer: PrintWriter, path: String, conn: Connection) =
        writeByJsonPath(tableName, writer, path, null, conn)

    /**
     * Retrieve the first document using a field comparison and optional ordering fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return The first JSON document matching the field comparison if found, an empty JSON object otherwise
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ): String {
        val named = Parameters.nameFields(fields)
        return Custom.jsonSingle(
            FindQuery.byFields(tableName, named, howMatched) + (orderBy?.let(::orderBy) ?: ""),
            Parameters.addFields(named),
            conn,
            Results::jsonFromData
        )
    }

    /**
     * Retrieve the first document using a field comparison and optional ordering fields (creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first JSON document matching the field comparison if found, an empty JSON object otherwise
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { firstByFields(tableName, fields, howMatched, orderBy, it) }

    /**
     * Retrieve the first document using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param conn The connection over which documents should be retrieved
     * @return The first JSON document matching the field comparison if found, an empty JSON object otherwise
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun firstByFields(
        tableName: String,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) = firstByFields(tableName, fields, howMatched, null, conn)

    /**
     * Write the first document to the given `PrintWriter` using a field comparison and optional ordering fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeFirstByFields(
        tableName: String,
        writer: PrintWriter,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = writer.write(firstByFields(tableName, fields, howMatched, orderBy, conn))

    /**
     * Write the first document to the given `PrintWriter` using a field comparison and optional ordering fields
     * (creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If no connection string has been set, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun writeFirstByFields(
        tableName: String,
        writer: PrintWriter,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { writeFirstByFields(tableName, writer, fields, howMatched, orderBy, it) }

    /**
     * Write the first document to the given `PrintWriter` using a field comparison
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeFirstByFields(
        tableName: String,
        writer: PrintWriter,
        fields: Collection<Field<*>>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) = writeFirstByFields(tableName, writer, fields, howMatched, null, conn)


    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return The first JSON document matching the JSON containment query if found, an empty JSON object otherwise
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> firstByContains(
        tableName: String,
        criteria: TContains,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = Custom.jsonSingle(
        FindQuery.byContains(tableName) + (orderBy?.let(::orderBy) ?: ""),
        listOf(Parameters.json(":criteria", criteria)),
        conn,
        Results::jsonFromData
    )

    /**
     * Retrieve the first document using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection over which documents should be retrieved
     * @return The first JSON document matching the JSON containment query if found, an empty JSON object otherwise
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> firstByContains(tableName: String, criteria: TContains, conn: Connection) =
        firstByContains(tableName, criteria, null, conn)

    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only; creates
     * connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first JSON document matching the JSON containment query if found, an empty JSON object otherwise
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TContains> firstByContains(tableName: String, criteria: TContains, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { firstByContains(tableName, criteria, orderBy, it) }

    /**
     * Write the first document to the given `PrintWriter` using a JSON containment query and optional ordering fields
     * (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> writeFirstByContains(
        tableName: String,
        writer: PrintWriter,
        criteria: TContains,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = writer.write(firstByContains(tableName, criteria, orderBy, conn))

    /**
     * Write the first document to the given `PrintWriter` using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun <TContains> writeFirstByContains(
        tableName: String,
        writer: PrintWriter,
        criteria: TContains,
        conn: Connection
    ) = writeFirstByContains(tableName, writer, criteria, null, conn)

    /**
     * Write the first document to the given `PrintWriter` using a JSON containment query and optional ordering fields
     * (PostgreSQL only; creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param criteria The object for which JSON containment should be checked
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TContains> writeFirstByContains(
        tableName: String,
        writer: PrintWriter,
        criteria: TContains,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { writeFirstByContains(tableName, writer, criteria, orderBy, it) }

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @return The first JSON document matching the JSON Path match query if found, an empty JSON object otherwise
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun firstByJsonPath(tableName: String, path: String, orderBy: Collection<Field<*>>? = null, conn: Connection) =
        Custom.jsonSingle(
            FindQuery.byJsonPath(tableName) + (orderBy?.let(::orderBy) ?: ""),
            listOf(Parameter(":path", ParameterType.STRING, path)),
            conn,
            Results::jsonFromData
        )

    /**
     * Retrieve the first document using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param conn The connection over which documents should be retrieved
     * @return The first JSON document matching the JSON Path match query if found, an empty JSON object otherwise
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun firstByJsonPath(tableName: String, path: String, conn: Connection) =
        firstByJsonPath(tableName, path, null, conn)

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only; creates
     * connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first JSON document matching the JSON Path match query if found, an empty JSON object otherwise
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun firstByJsonPath(tableName: String, path: String, orderBy: Collection<Field<*>>? = null) =
        Configuration.dbConn().use { firstByJsonPath(tableName, path, orderBy, it) }

    /**
     * Write the first document to the given `PrintWriter` using a JSON Path match query and optional ordering fields
     * (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeFirstByJsonPath(
        tableName: String,
        writer: PrintWriter,
        path: String,
        orderBy: Collection<Field<*>>? = null,
        conn: Connection
    ) = writer.write(firstByJsonPath(tableName, path, orderBy, conn))

    /**
     * Write the first document to the given `PrintWriter` using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param path The JSON path comparison to match
     * @param conn The connection over which documents should be retrieved
     * @throws DocumentException If called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun writeFirstByJsonPath(tableName: String, writer: PrintWriter, path: String, conn: Connection) =
        writeFirstByJsonPath(tableName, writer, path, null, conn)

    /**
     * Write the first document to the given `PrintWriter` using a JSON Path match query and optional ordering fields
     * (PostgreSQL only; creates connection)
     *
     * @param tableName The table from which documents should be retrieved
     * @param writer The `PrintWriter` to which the results should be written
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun writeFirstByJsonPath(
        tableName: String,
        writer: PrintWriter,
        path: String,
        orderBy: Collection<Field<*>>? = null
    ) = Configuration.dbConn().use { writeFirstByJsonPath(tableName, writer, path, orderBy, it) }
}
