package solutions.bitbadger.documents.core.tests.java;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.Dialect;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.DocumentIndex;
import solutions.bitbadger.documents.query.DefinitionQuery;
import solutions.bitbadger.documents.core.tests.ForceDialect;

import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;

/**
 * Unit tests for the `Definition` object
 */
@DisplayName("Core | Java | Query | DefinitionQuery")
final public class DefinitionQueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    public void cleanUp() {
        ForceDialect.none();
    }

    @Test
    @DisplayName("ensureTableFor generates correctly")
    public void ensureTableFor() {
        assertEquals("CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)",
                DefinitionQuery.ensureTableFor("my.table", "JSONB"),
                "CREATE TABLE statement not constructed correctly");
    }

    @Test
    @DisplayName("ensureTable generates correctly | PostgreSQL")
    public void ensureTablePostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("CREATE TABLE IF NOT EXISTS %s (data JSONB NOT NULL)", TEST_TABLE),
                DefinitionQuery.ensureTable(TEST_TABLE));
    }

    @Test
    @DisplayName("ensureTable generates correctly | SQLite")
    public void ensureTableSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals(String.format("CREATE TABLE IF NOT EXISTS %s (data TEXT NOT NULL)", TEST_TABLE),
                DefinitionQuery.ensureTable(TEST_TABLE));
    }

    @Test
    @DisplayName("ensureTable fails when no dialect is set")
    public void ensureTableFailsUnknown() {
        assertThrows(DocumentException.class, () -> DefinitionQuery.ensureTable(TEST_TABLE));
    }

    @Test
    @DisplayName("ensureKey generates correctly with schema")
    public void ensureKeyWithSchema() throws DocumentException {
        assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
                DefinitionQuery.ensureKey("test.table", Dialect.POSTGRESQL),
                "CREATE INDEX for key statement with schema not constructed correctly");
    }

    @Test
    @DisplayName("ensureKey generates correctly without schema")
    public void ensureKeyWithoutSchema() throws DocumentException {
        assertEquals(
                String.format("CREATE UNIQUE INDEX IF NOT EXISTS idx_%1$s_key ON %1$s ((data->>'id'))", TEST_TABLE),
                DefinitionQuery.ensureKey(TEST_TABLE, Dialect.SQLITE),
                "CREATE INDEX for key statement without schema not constructed correctly");
    }

    @Test
    @DisplayName("ensureIndexOn generates multiple fields and directions")
    public void ensureIndexOnMultipleFields() throws DocumentException {
        assertEquals(
                "CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table ((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
                DefinitionQuery.ensureIndexOn("test.table", "gibberish", List.of("taco", "guac DESC", "salsa ASC"),
                        Dialect.POSTGRESQL),
                "CREATE INDEX for multiple field statement not constructed correctly");
    }

    @Test
    @DisplayName("ensureIndexOn generates nested field | PostgreSQL")
    public void ensureIndexOnNestedPostgres() throws DocumentException {
        assertEquals(String.format("CREATE INDEX IF NOT EXISTS idx_%1$s_nest ON %1$s ((data#>>'{a,b,c}'))", TEST_TABLE),
                DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", List.of("a.b.c"), Dialect.POSTGRESQL),
                "CREATE INDEX for nested PostgreSQL field incorrect");
    }

    @Test
    @DisplayName("ensureIndexOn generates nested field | SQLite")
    public void ensureIndexOnNestedSQLite() throws DocumentException {
        assertEquals(
                String.format("CREATE INDEX IF NOT EXISTS idx_%1$s_nest ON %1$s ((data->'a'->'b'->>'c'))", TEST_TABLE),
                DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", List.of("a.b.c"), Dialect.SQLITE),
                "CREATE INDEX for nested SQLite field incorrect");
    }

    @Test
    @DisplayName("ensureDocumentIndexOn generates Full | PostgreSQL")
    public void ensureDocumentIndexOnFullPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("CREATE INDEX IF NOT EXISTS idx_%1$s_document ON %1$s USING GIN (data)", TEST_TABLE),
                DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL),
                "CREATE INDEX for full document index incorrect");
    }

    @Test
    @DisplayName("ensureDocumentIndexOn generates Optimized | PostgreSQL")
    public void ensureDocumentIndexOnOptimizedPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(
                String.format("CREATE INDEX IF NOT EXISTS idx_%1$s_document ON %1$s USING GIN (data jsonb_path_ops)",
                        TEST_TABLE),
                DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.OPTIMIZED),
                "CREATE INDEX for optimized document index incorrect");
    }

    @Test
    @DisplayName("ensureDocumentIndexOn fails | SQLite")
    public void ensureDocumentIndexOnFailsSQLite() {
        ForceDialect.sqlite();
        assertThrows(DocumentException.class,
                () -> DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL));
    }
}
