package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.Configuration;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.FieldMatch;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;

import java.util.List;
import java.util.Map;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;

final public class FindFunctions {

    private static String docIds(List<JsonDocument> docs) {
        return docs.stream().map(JsonDocument::getId).reduce((acc, docId) -> String.format("%s|%s", acc, docId)).get();
    }

    public static void allDefault(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(5, findAll(db.getConn(), TEST_TABLE, JsonDocument.class).size(),
                "There should have been 5 documents returned");
    }

    public static void allAscending(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findAll(db.getConn(), TEST_TABLE, JsonDocument.class,
                List.of(Field.named("id")));
        assertEquals(5, docs.size(), "There should have been 5 documents returned");
        assertEquals("five|four|one|three|two", docIds(docs), "The documents were not ordered correctly");
    }

    public static void allDescending(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findAll(db.getConn(), TEST_TABLE, JsonDocument.class,
                List.of(Field.named("id DESC")));
        assertEquals(5, docs.size(), "There should have been 5 documents returned");
        assertEquals("two|three|one|four|five", docIds(docs), "The documents were not ordered correctly");
    }

    public static void allNumOrder(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findAll(db.getConn(), TEST_TABLE, JsonDocument.class,
                List.of(Field.named("sub.foo NULLS LAST"), Field.named("n:numValue")));
        assertEquals(5, docs.size(), "There should have been 5 documents returned");
        assertEquals("two|four|one|three|five", docIds(docs), "The documents were not ordered correctly");
    }

    public static void allEmpty(ThrowawayDatabase db) throws DocumentException {
        assertEquals(0, findAll(db.getConn(), TEST_TABLE, JsonDocument.class).size(),
                "There should have been no documents returned");
    }

    public static void byIdString(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findById(db.getConn(), TEST_TABLE, "two", JsonDocument.class);
        assertTrue(doc.isPresent(), "The document should have been returned");
        assertEquals("two", doc.get().getId(), "An incorrect document was returned");
    }

    public static void byIdNumber(ThrowawayDatabase db) throws DocumentException {
        Configuration.idField = "key";
        try {
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(18, "howdy"));
            final Optional<NumIdDocument> doc = findById(db.getConn(), TEST_TABLE, 18, NumIdDocument.class);
            assertTrue(doc.isPresent(), "The document should have been returned");
        } finally {
            Configuration.idField = "id";
        }
    }

    public static void byIdNotFound(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(findById(db.getConn(), TEST_TABLE, "x", JsonDocument.class).isPresent(),
                "There should have been no document returned");
    }

    public static void byFieldsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByFields(db.getConn(), TEST_TABLE,
                List.of(Field.any("value", List.of("blue", "purple")), Field.exists("sub")), JsonDocument.class,
                FieldMatch.ALL);
        assertEquals(1, docs.size(), "There should have been a document returned");
        assertEquals("four", docs.get(0).getId(), "The incorrect document was returned");
    }

    public static void byFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "purple")),
                JsonDocument.class, null, List.of(Field.named("id")));
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        assertEquals("five|four", docIds(docs), "The documents were not ordered correctly");
    }

    public static void byFieldsMatchNumIn(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByFields(db.getConn(), TEST_TABLE,
                List.of(Field.any("numValue", List.of(2, 4, 6, 8))), JsonDocument.class);
        assertEquals(1, docs.size(), "There should have been a document returned");
        assertEquals("three", docs.get(0).getId(), "The incorrect document was returned");
    }

    public static void byFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(0,
                findByFields(db.getConn(), TEST_TABLE, List.of(Field.greater("numValue", 100)), JsonDocument.class)
                        .size(),
                "There should have been no documents returned");
    }

    public static void byFieldsMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (final ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        final List<ArrayDocument> docs = findByFields(db.getConn(), TEST_TABLE,
                List.of(Field.inArray("values", TEST_TABLE, List.of("c"))), ArrayDocument.class);
        assertEquals(2, docs.size(), "There should have been two documents returned");
        assertTrue(List.of("first", "second").contains(docs.get(0).getId()),
                String.format("An incorrect document was returned (%s)", docs.get(0).getId()));
        assertTrue(List.of("first", "second").contains(docs.get(1).getId()),
                String.format("An incorrect document was returned (%s)", docs.get(1).getId()));
    }

    public static void byFieldsNoMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (final ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        assertEquals(0,
                findByFields(db.getConn(), TEST_TABLE, List.of(Field.inArray("values", TEST_TABLE, List.of("j"))),
                        ArrayDocument.class).size(),
                "There should have been no documents returned");
    }

    public static void byContainsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"),
                JsonDocument.class);
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        assertTrue(List.of("four", "five").contains(docs.get(0).getId()),
                String.format("An incorrect document was returned (%s)", docs.get(0).getId()));
        assertTrue(List.of("four", "five").contains(docs.get(1).getId()),
                String.format("An incorrect document was returned (%s)", docs.get(1).getId()));
    }

    public static void byContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByContains(db.getConn(), TEST_TABLE, Map.of("sub", Map.of("foo", "green")),
                JsonDocument.class, List.of(Field.named("value")));
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        assertEquals("two|four", docIds(docs), "The documents were not ordered correctly");
    }

    public static void byContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(0, findByContains(db.getConn(), TEST_TABLE, Map.of("value", "indigo"), JsonDocument.class).size(),
                "There should have been no documents returned");
    }

    public static void byJsonPathMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                JsonDocument.class);
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        assertTrue(List.of("four", "five").contains(docs.get(0).getId()),
                String.format("An incorrect document was returned (%s)", docs.get(0).getId()));
        assertTrue(List.of("four", "five").contains(docs.get(1).getId()),
                String.format("An incorrect document was returned (%s)", docs.get(1).getId()));    }

    public static void byJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<JsonDocument> docs = findByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                JsonDocument.class, List.of(Field.named("id")));
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        assertEquals("five|four", docIds(docs), "The documents were not ordered correctly");
    }

    public static void byJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(0, findByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 100)", JsonDocument.class).size(),
                "There should have been no documents returned");
    }

    public static void firstByFieldsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByFields(db.getConn(), TEST_TABLE,
                List.of(Field.equal("value", "another")), JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("two", doc.get().getId(), "The incorrect document was returned");
    }

    public static void firstByFieldsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByFields(db.getConn(), TEST_TABLE,
                List.of(Field.equal("sub.foo", "green")), JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertTrue(List.of("two", "four").contains(doc.get().getId()),
                String.format("An incorrect document was returned (%s)", doc.get().getId()));
    }

    public static void firstByFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByFields(db.getConn(), TEST_TABLE,
                List.of(Field.equal("sub.foo", "green")), JsonDocument.class, null,
                List.of(Field.named("n:numValue DESC")));
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("four", doc.get().getId(), "An incorrect document was returned");
    }

    public static void firstByFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(findFirstByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "absent")),
                        JsonDocument.class).isPresent(),
                "There should have been no document returned");
    }

    public static void firstByContainsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "FIRST!"),
                JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("one", doc.get().getId(), "An incorrect document was returned");
    }

    public static void firstByContainsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"),
                JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertTrue(List.of("four", "five").contains(doc.get().getId()),
                String.format("An incorrect document was returned (%s)", doc.get().getId()));
    }

    public static void firstByContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"),
                JsonDocument.class, List.of(Field.named("sub.bar NULLS FIRST")));
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("five", doc.get().getId(), "An incorrect document was returned");
    }

    public static void firstByContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(findFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "indigo"), JsonDocument.class)
                        .isPresent(),
                "There should have been no document returned");
    }

    public static void firstByJsonPathMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ == 10)",
                JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("two", doc.get().getId(), "An incorrect document was returned");
    }

    public static void firstByJsonPathMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertTrue(List.of("four", "five").contains(doc.get().getId()),
                String.format("An incorrect document was returned (%s)", doc.get().getId()));
    }

    public static void firstByJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Optional<JsonDocument> doc = findFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                JsonDocument.class, List.of(Field.named("id DESC")));
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("four", doc.get().getId(), "An incorrect document was returned");
    }

    public static void firstByJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(findFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 100)", JsonDocument.class)
                        .isPresent(),
                "There should have been no document returned");
    }
}
