package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.*;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.List;
import java.util.Map;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;

/**
 * Tests for the JSON-returning functions
 * <p>
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
final public class JsonFunctions {

    /**
     * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
     * This function will do a crude string replacement to match the target string based on the dialect being tested.
     *
     * @param json The JSON which should be returned
     * @return The actual expected JSON based on the database being tested
     */
    public static String maybeJsonB(String json) throws DocumentException {
        return switch (Configuration.dialect()) {
            case SQLITE -> json;
            case POSTGRESQL -> json.replace("\":", "\": ").replace(",\"", ", \"");
        };
    }

    /**
     * Create a snippet of JSON to find a document ID
     *
     * @param id The ID of the document
     * @return A connection-aware ID to check for presence and positioning
     */
    private static String docId(String id) throws DocumentException {
        return maybeJsonB(String.format("{\"id\":\"%s\"", id));
    }

    private static void checkAllDefault(String json) throws DocumentException {
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)");
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.one),
                        String.format("Document 'one' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.two),
                        String.format("Document 'two' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.three),
                        String.format("Document 'three' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.four),
                        String.format("Document 'four' not found in JSON (%s)", json));
                assertTrue(json.contains(JsonDocument.five),
                        String.format("Document 'five' not found in JSON (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("one")), String.format("Document 'one' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("two")), String.format("Document 'two' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("three")),
                        String.format("Document 'three' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("four")), String.format("Document 'four' not found in JSON (%s)", json));
                assertTrue(json.contains(docId("five")), String.format("Document 'five' not found in JSON (%s)", json));
                break;
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)");
    }

    public static void allDefault(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkAllDefault(jsonAll(db.getConn(), TEST_TABLE));
    }

    public static void writeAllDefault(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonAll(db.getConn(), TEST_TABLE, writer);
        checkAllDefault(output.toString());
    }

    private static void checkAllEmpty(String json) {
        assertEquals("[]", json, "There should have been no documents returned");
    }

    public static void allEmpty(ThrowawayDatabase db) throws DocumentException {
        checkAllEmpty(jsonAll(db.getConn(), TEST_TABLE));
    }

    public static void writeAllEmpty(ThrowawayDatabase db) throws DocumentException {
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonAll(db.getConn(), TEST_TABLE, writer);
        checkAllEmpty(output.toString());
    }

    private static void checkByIdString(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.two, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")), String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void byIdString(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByIdString(jsonById(db.getConn(), TEST_TABLE, "two"));
    }

    public static void writeByIdString(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonById(db.getConn(), TEST_TABLE, writer, "two");
        checkByIdString(output.toString());
    }

    private static void checkByIdNumber(String json) throws DocumentException {
        assertEquals(maybeJsonB("{\"key\":18,\"text\":\"howdy\"}"), json,
                "The document should have been found by numeric ID");
    }

    public static void byIdNumber(ThrowawayDatabase db) throws DocumentException {
        Configuration.idField = "key";
        try {
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(18, "howdy"));
            checkByIdNumber(jsonById(db.getConn(), TEST_TABLE, 18));
        } finally {
            Configuration.idField = "id";
        }
    }

    public static void writeByIdNumber(ThrowawayDatabase db) throws DocumentException {
        Configuration.idField = "key";
        try {
            insert(db.getConn(), TEST_TABLE, new NumIdDocument(18, "howdy"));
            final StringWriter output = new StringWriter();
            final PrintWriter writer = new PrintWriter(output);
            writeJsonById(db.getConn(), TEST_TABLE, writer, 18);
            checkByIdNumber(output.toString());
        } finally {
            Configuration.idField = "id";
        }
    }

    private static void checkByIdNotFound(String json) {
        assertEquals("{}", json, "There should have been no document returned");
    }

    public static void byIdNotFound(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByIdNotFound(jsonById(db.getConn(), TEST_TABLE, "x"));
    }

    public static void writeByIdNotFound(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonById(db.getConn(), TEST_TABLE, writer, "x");
        checkByIdNotFound(output.toString());
    }

    private static void checkByFieldsMatch(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s]", JsonDocument.four), json, "The incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(json.contains(docId("four")),
                        String.format("The incorrect document was returned (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byFieldsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByFieldsMatch(jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.any("value", List.of("blue", "purple")),
                Field.exists("sub")), FieldMatch.ALL));
    }

    public static void writeByFieldsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.any("value", List.of("blue", "purple")),
                Field.exists("sub")), FieldMatch.ALL);
        checkByFieldsMatch(output.toString());
    }

    private static void checkByFieldsMatchOrdered(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s,%s]", JsonDocument.five, JsonDocument.four), json,
                        "The documents were not ordered correctly");
                break;
            case POSTGRESQL:
                final int fiveIdx = json.indexOf(docId("five"));
                final int fourIdx = json.indexOf(docId("four"));
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(fiveIdx >= 0, String.format("Document 'five' not found (%s)", json));
                assertTrue(fourIdx >= 0, String.format("Document 'four' not found (%s)", json));
                assertTrue(fiveIdx < fourIdx,
                        String.format("Document 'five' should have been before 'four' (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByFieldsMatchOrdered(jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "purple")), null,
                List.of(Field.named("id"))));
    }

    public static void writeByFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.equal("value", "purple")), null,
                List.of(Field.named("id")));
        checkByFieldsMatchOrdered(output.toString());
    }

    private static void checkByFieldsMatchNumIn(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s]", JsonDocument.three), json, "The incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(json.contains(docId("three")),
                        String.format("The incorrect document was returned (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byFieldsMatchNumIn(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByFieldsMatchNumIn(jsonByFields(db.getConn(), TEST_TABLE,
                List.of(Field.any("numValue", List.of(2, 4, 6, 8)))));
    }

    public static void writeByFieldsMatchNumIn(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.any("numValue", List.of(2, 4, 6, 8))));
        checkByFieldsMatchNumIn(output.toString());
    }

    private static void checkByFieldsNoMatch(String json) {
        assertEquals("[]", json, "There should have been no documents returned");
    }

    public static void byFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByFieldsNoMatch(jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.greater("numValue", 100))));
    }

    public static void writeByFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.greater("numValue", 100)));
        checkByFieldsNoMatch(output.toString());
    }

    private static void checkByFieldsMatchInArray(String json) throws DocumentException {
        assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
        assertTrue(json.contains(docId("first")), String.format("The 'first' document was not found (%s)", json));
        assertTrue(json.contains(docId("second")), String.format("The 'second' document was not found (%s)", json));
        assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
    }

    public static void byFieldsMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        checkByFieldsMatchInArray(jsonByFields(db.getConn(), TEST_TABLE, List.of(Field.inArray("values", TEST_TABLE,
                List.of("c")))));
    }

    public static void writeByFieldsMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.inArray("values", TEST_TABLE, List.of("c"))));
        checkByFieldsMatchInArray(output.toString());
    }

    private static void checkByFieldsNoMatchInArray(String json) {
        assertEquals("[]", json, "There should have been no documents returned");
    }

    public static void byFieldsNoMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        checkByFieldsNoMatchInArray(jsonByFields(db.getConn(), TEST_TABLE,
                List.of(Field.inArray("values", TEST_TABLE, List.of("j")))));
    }

    public static void writeByFieldsNoMatchInArray(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.inArray("values", TEST_TABLE, List.of("j"))));
        checkByFieldsNoMatchInArray(output.toString());
    }

    private static void checkByContainsMatch(String json) throws DocumentException {
        assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(JsonDocument.five), String.format("Document 'five' not found (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(docId("five")), String.format("Document 'five' not found (%s)", json));
                break;
        }
        assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
    }

    public static void byContainsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByContainsMatch(jsonByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple")));
    }

    public static void writeByContainsMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByContains(db.getConn(), TEST_TABLE, writer, Map.of("value", "purple"));
        checkByContainsMatch(output.toString());
    }

    private static void checkByContainsMatchOrdered(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s,%s]", JsonDocument.two, JsonDocument.four), json,
                        "The documents were not ordered correctly");
                break;
            case POSTGRESQL:
                final int twoIdx = json.indexOf(docId("two"));
                final int fourIdx = json.indexOf(docId("four"));
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(twoIdx >= 0, String.format("Document 'two' not found (%s)", json));
                assertTrue(fourIdx >= 0, String.format("Document 'four' not found (%s)", json));
                assertTrue(twoIdx < fourIdx, String.format("Document 'two' should have been before 'four' (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByContainsMatchOrdered(jsonByContains(db.getConn(), TEST_TABLE, Map.of("sub", Map.of("foo", "green")),
                List.of(Field.named("value"))));
    }

    public static void writeByContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByContains(db.getConn(), TEST_TABLE, writer, Map.of("sub", Map.of("foo", "green")),
                List.of(Field.named("value")));
        checkByContainsMatchOrdered(output.toString());
    }

    private static void checkByContainsNoMatch(String json) {
        assertEquals("[]", json, "There should have been no documents returned");
    }

    public static void byContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByContainsNoMatch(jsonByContains(db.getConn(), TEST_TABLE, Map.of("value", "indigo")));
    }

    public static void writeByContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByContains(db.getConn(), TEST_TABLE, writer, Map.of("value", "indigo"));
        checkByContainsNoMatch(output.toString());
    }

    private static void checkByJsonPathMatch(String json) throws DocumentException {
        assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(JsonDocument.five), String.format("Document 'five' not found (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")), String.format("Document 'four' not found (%s)", json));
                assertTrue(json.contains(docId("five")), String.format("Document 'five' not found (%s)", json));
                break;
        }
        assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
    }

    public static void byJsonPathMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByJsonPathMatch(jsonByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)"));
    }

    public static void writeByJsonPathMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ > 10)");
        checkByJsonPathMatch(output.toString());
    }

    private static void checkByJsonPathMatchOrdered(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(String.format("[%s,%s]", JsonDocument.five, JsonDocument.four), json,
                        "The documents were not ordered correctly");
                break;
            case POSTGRESQL:
                final int fiveIdx = json.indexOf(docId("five"));
                final int fourIdx = json.indexOf(docId("four"));
                assertTrue(json.startsWith("["), String.format("JSON should start with '[' (%s)", json));
                assertTrue(fiveIdx >= 0, String.format("Document 'five' not found (%s)", json));
                assertTrue(fourIdx >= 0, String.format("Document 'four' not found (%s)", json));
                assertTrue(fiveIdx < fourIdx,
                        String.format("Document 'five' should have been before 'four' (%s)", json));
                assertTrue(json.endsWith("]"), String.format("JSON should end with ']' (%s)", json));
                break;
        }
    }

    public static void byJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByJsonPathMatchOrdered(jsonByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                List.of(Field.named("id"))));
    }

    public static void writeByJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ > 10)", List.of(Field.named("id")));
        checkByJsonPathMatchOrdered(output.toString());
    }

    private static void checkByJsonPathNoMatch(String json) {
        assertEquals("[]", json, "There should have been no documents returned");
    }

    public static void byJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkByJsonPathNoMatch(jsonByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 100)"));
    }

    public static void writeByJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ > 100)");
        checkByJsonPathNoMatch(output.toString());
    }

    private static void checkFirstByFieldsMatchOne(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.two, json, "The incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")),
                        String.format("The incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByFieldsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByFieldsMatchOne(jsonFirstByFields(db.getConn(), TEST_TABLE,
                List.of(Field.equal("value", "another"))));
    }

    public static void writeFirstByFieldsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.equal("value", "another")));
        checkFirstByFieldsMatchOne(output.toString());
    }

    private static void checkFirstByFieldsMatchMany(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
                        String.format("Expected document 'two' or 'four' (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")) || json.contains(docId("four")),
                        String.format("Expected document 'two' or 'four' (%s)", json));
                break;
        }
    }

    public static void firstByFieldsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByFieldsMatchMany(jsonFirstByFields(db.getConn(), TEST_TABLE,
                List.of(Field.equal("sub.foo", "green"))));
    }

    public static void writeFirstByFieldsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.equal("sub.foo", "green")));
        checkFirstByFieldsMatchMany(output.toString());
    }

    private static void checkFirstByFieldsMatchOrdered(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.four, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")),
                        String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByFieldsMatchOrdered(jsonFirstByFields(db.getConn(), TEST_TABLE,
                List.of(Field.equal("sub.foo", "green")), null, List.of(Field.named("n:numValue DESC"))));
    }

    public static void writeFirstByFieldsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.equal("sub.foo", "green")), null,
                List.of(Field.named("n:numValue DESC")));
        checkFirstByFieldsMatchOrdered(output.toString());
    }

    private static void checkFirstByFieldsNoMatch(String json) {
        assertEquals("{}", json, "There should have been no document returned");
    }

    public static void firstByFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByFieldsNoMatch(jsonFirstByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("value", "absent"))));
    }

    public static void writeFirstByFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByFields(db.getConn(), TEST_TABLE, writer, List.of(Field.equal("value", "absent")));
        checkFirstByFieldsNoMatch(output.toString());
    }

    private static void checkFirstByContainsMatchOne(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.one, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("one")), String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByContainsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByContainsMatchOne(jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "FIRST!")));
    }

    public static void writeFirstByContainsMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByContains(db.getConn(), TEST_TABLE, writer, Map.of("value", "FIRST!"));
        checkFirstByContainsMatchOne(output.toString());
    }

    private static void checkFirstByContainsMatchMany(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
        }
    }

    public static void firstByContainsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByContainsMatchMany(jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple")));
    }

    public static void writeFirstByContainsMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByContains(db.getConn(), TEST_TABLE, writer, Map.of("value", "purple"));
        checkFirstByContainsMatchMany(output.toString());
    }

    private static void checkFirstByContainsMatchOrdered(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.five, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("five")),
                        String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByContainsMatchOrdered(jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "purple"),
                List.of(Field.named("sub.bar NULLS FIRST"))));
    }

    public static void writeFirstByContainsMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByContains(db.getConn(), TEST_TABLE, writer, Map.of("value", "purple"),
                List.of(Field.named("sub.bar NULLS FIRST")));
        checkFirstByContainsMatchOrdered(output.toString());
    }

    private static void checkFirstByContainsNoMatch(String json) {
        assertEquals("{}", json, "There should have been no document returned");
    }

    public static void firstByContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByContainsNoMatch(jsonFirstByContains(db.getConn(), TEST_TABLE, Map.of("value", "indigo")));
    }

    public static void writeFirstByContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByContains(db.getConn(), TEST_TABLE, writer, Map.of("value", "indigo"));
        checkFirstByContainsNoMatch(output.toString());
    }

    private static void checkFirstByJsonPathMatchOne(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.two, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("two")), String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByJsonPathMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByJsonPathMatchOne(jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ == 10)"));
    }

    public static void writeFirstByJsonPathMatchOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ == 10)");
        checkFirstByJsonPathMatchOne(output.toString());
    }

    private static void checkFirstByJsonPathMatchMany(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
                        String.format("Expected document 'four' or 'five' (%s)", json));
                break;
        }
    }

    public static void firstByJsonPathMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByJsonPathMatchMany(jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)"));
    }

    public static void writeFirstByJsonPathMatchMany(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ > 10)");
        checkFirstByJsonPathMatchMany(output.toString());
    }

    private static void checkFirstByJsonPathMatchOrdered(String json) throws DocumentException {
        switch (Configuration.dialect()) {
            case SQLITE:
                assertEquals(JsonDocument.four, json, "An incorrect document was returned");
                break;
            case POSTGRESQL:
                assertTrue(json.contains(docId("four")),
                        String.format("An incorrect document was returned (%s)", json));
                break;
        }
    }

    public static void firstByJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByJsonPathMatchOrdered(jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 10)",
                List.of(Field.named("id DESC"))));
    }

    public static void writeFirstByJsonPathMatchOrdered(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ > 10)",
                List.of(Field.named("id DESC")));
        checkFirstByJsonPathMatchOrdered(output.toString());
    }

    private static void checkFirstByJsonPathNoMatch(String json) {
        assertEquals("{}", json, "There should have been no document returned");
    }

    public static void firstByJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        checkFirstByJsonPathNoMatch(jsonFirstByJsonPath(db.getConn(), TEST_TABLE, "$.numValue ? (@ > 100)"));
    }

    public static void writeFirstByJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeJsonFirstByJsonPath(db.getConn(), TEST_TABLE, writer, "$.numValue ? (@ > 100)");
        checkFirstByJsonPathNoMatch(output.toString());
    }
}
