package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;

import java.util.List;
import java.util.Map;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;

final public class RemoveFieldsFunctions {

    public static void byIdMatchFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        removeFieldsById(db.getConn(), TEST_TABLE, "two", List.of("sub", "value"));
        final Optional<JsonDocument> doc = findById(db.getConn(), TEST_TABLE, "two", JsonDocument.class);
        assertTrue(doc.isPresent(), "There should have been a document returned");
        assertEquals("", doc.get().getValue(), "The value should have been empty");
        assertNull(doc.get().getSub(), "The sub-document should have been removed");
    }

    public static void byIdMatchNoFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsByFields(db.getConn(), TEST_TABLE, List.of(Field.exists("a_field_that_does_not_exist"))));
        removeFieldsById(db.getConn(), TEST_TABLE, "one", List.of("a_field_that_does_not_exist")); // no exn = pass
    }

    public static void byIdNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsById(db.getConn(), TEST_TABLE, "fifty"));
        removeFieldsById(db.getConn(), TEST_TABLE, "fifty", List.of("sub")); // no exception = pass
    }

    public static void byFieldsMatchFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<Field<?>> fields = List.of(Field.equal("numValue", 17));
        removeFieldsByFields(db.getConn(), TEST_TABLE, fields, List.of("sub"));
        final Optional<JsonDocument> doc = findFirstByFields(db.getConn(), TEST_TABLE, fields, JsonDocument.class);
        assertTrue(doc.isPresent(), "The document should have been returned");
        assertEquals("four", doc.get().getId(), "An incorrect document was returned");
        assertNull(doc.get().getSub(), "The sub-document should have been removed");
    }

    public static void byFieldsMatchNoFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsByFields(db.getConn(), TEST_TABLE, List.of(Field.exists("nada"))));
        removeFieldsByFields(db.getConn(), TEST_TABLE, List.of(Field.equal("numValue", 17)), List.of("nada"));
        // no exception = pass
    }

    public static void byFieldsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final List<Field<?>> fields = List.of(Field.notEqual("missing", "nope"));
        assertFalse(existsByFields(db.getConn(), TEST_TABLE, fields));
        removeFieldsByFields(db.getConn(), TEST_TABLE, fields, List.of("value")); // no exception = pass
    }

    public static void byContainsMatchFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Map<String, Map<String, String>> criteria = Map.of("sub", Map.of("foo", "green"));
        removeFieldsByContains(db.getConn(), TEST_TABLE, criteria, List.of("value"));
        final List<JsonDocument> docs = findByContains(db.getConn(), TEST_TABLE, criteria, JsonDocument.class);
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        for (final JsonDocument doc : docs) {
            assertTrue(List.of("two", "four").contains(doc.getId()),
                    String.format("An incorrect document was returned (%s)", doc.getId()));
            assertEquals("", doc.getValue(), "The value should have been empty");
        }
    }

    public static void byContainsMatchNoFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsByFields(db.getConn(), TEST_TABLE, List.of(Field.exists("invalid_field"))));
        removeFieldsByContains(db.getConn(), TEST_TABLE, Map.of("sub", Map.of("foo", "green")),
                List.of("invalid_field")); // no exception = pass
    }

    public static void byContainsNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final Map<String, String> contains = Map.of("value", "substantial");
        assertFalse(existsByContains(db.getConn(), TEST_TABLE, contains));
        removeFieldsByContains(db.getConn(), TEST_TABLE, contains, List.of("numValue")); // no exception = pass
    }

    public static void byJsonPathMatchFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String path = "$.value ? (@ == \"purple\")";
        removeFieldsByJsonPath(db.getConn(), TEST_TABLE, path, List.of("sub"));
        final List<JsonDocument> docs = findByJsonPath(db.getConn(), TEST_TABLE, path, JsonDocument.class);
        assertEquals(2, docs.size(), "There should have been 2 documents returned");
        for (final JsonDocument doc : docs) {
            assertTrue(List.of("four", "five").contains(doc.getId()),
                    String.format("An incorrect document was returned (%s)", doc.getId()));
            assertNull(doc.getSub(), "The sub-document should have been removed");
        }
    }

    public static void byJsonPathMatchNoFields(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertFalse(existsByFields(db.getConn(), TEST_TABLE, List.of(Field.exists("submarine"))));
        removeFieldsByJsonPath(db.getConn(), TEST_TABLE, "$.value ? (@ == \"purple\")", List.of("submarine"));
        // no exception = pass
    }

    public static void byJsonPathNoMatch(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        final String path = "$.value ? (@ == \"mauve\")";
        assertFalse(existsByJsonPath(db.getConn(), TEST_TABLE, path));
        removeFieldsByJsonPath(db.getConn(), TEST_TABLE, path, List.of("value")); // no exception = pass
    }
}
