package solutions.bitbadger.documents.core.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.query.RemoveFieldsQuery
import kotlin.test.assertEquals

/**
 * Unit tests for the `RemoveFields` object
 */
@DisplayName("Core | Kotlin | Query | RemoveFieldsQuery")
class RemoveFieldsQueryTest {

    /**
     * Reset the dialect
     */
    @AfterEach
    fun cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName("byId generates correctly | PostgreSQL")
    fun byIdPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE data->>'id' = :id",
            RemoveFieldsQuery.byId<String>(TEST_TABLE, listOf(Parameter(":name", ParameterType.STRING, "{a,z}"))),
            "Remove Fields query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byId generates correctly | SQLite")
    fun byIdSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "UPDATE $TEST_TABLE SET data = json_remove(data, :name0, :name1) WHERE data->>'id' = :id",
            RemoveFieldsQuery.byId<String>(
                TEST_TABLE,
                listOf(
                    Parameter(":name0", ParameterType.STRING, "a"),
                    Parameter(":name1", ParameterType.STRING, "z")
                )
            ),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byFields generates correctly | PostgreSQL")
    fun byFieldsPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE data->>'f' > :g",
            RemoveFieldsQuery.byFields(
                TEST_TABLE,
                listOf(Parameter(":name", ParameterType.STRING, "{b,c}")),
                listOf(Field.greater("f", "", ":g"))
            ),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byFields generates correctly | SQLite")
    fun byFieldsSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "UPDATE $TEST_TABLE SET data = json_remove(data, :name0, :name1) WHERE data->>'f' > :g",
            RemoveFieldsQuery.byFields(
                TEST_TABLE,
                listOf(Parameter(":name0", ParameterType.STRING, "b"), Parameter(":name1", ParameterType.STRING, "c")),
                listOf(Field.greater("f", "", ":g"))
            ),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byContains generates correctly | PostgreSQL")
    fun byContainsPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE data @> :criteria",
            RemoveFieldsQuery.byContains(TEST_TABLE, listOf(Parameter(":name", ParameterType.STRING, "{m,n}"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byContains fails | SQLite")
    fun byContainsSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { RemoveFieldsQuery.byContains(TEST_TABLE, listOf()) }
    }

    @Test
    @DisplayName("byJsonPath generates correctly | PostgreSQL")
    fun byJsonPathPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE jsonb_path_exists(data, :path::jsonpath)",
            RemoveFieldsQuery.byJsonPath(TEST_TABLE, listOf(Parameter(":name", ParameterType.STRING, "{o,p}"))),
            "Remove Field query not constructed correctly"
        )
    }

    @Test
    @DisplayName("byJsonPath fails | SQLite")
    fun byJsonPathSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { RemoveFieldsQuery.byJsonPath(TEST_TABLE, listOf()) }
    }
}
