package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldFormat
import solutions.bitbadger.documents.Op

import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `Field` class
 */
@DisplayName('Groovy | Field')
class FieldTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    void cleanUp() {
        ForceDialect.none()
    }

    // ~~~ INSTANCE METHODS ~~~

    @Test
    @DisplayName('withParameterName fails for invalid name')
    void withParamNameFails() {
        assertThrows(DocumentException) {  Field.equal('it', '').withParameterName('2424') }
    }

    @Test
    @DisplayName('withParameterName works with colon prefix')
    void withParamNameColon() {
        def field = Field.equal('abc', '22').withQualifier 'me'
        def withParam = field.withParameterName ':test'
        assertNotSame field, withParam, 'A new Field instance should have been created'
        assertEquals field.name, withParam.name, 'Name should have been preserved'
        assertEquals field.comparison, withParam.comparison, 'Comparison should have been preserved'
        assertEquals ':test', withParam.parameterName, 'Parameter name not set correctly'
        assertEquals field.qualifier, withParam.qualifier, 'Qualifier should have been preserved'
    }

    @Test
    @DisplayName('withParameterName works with at-sign prefix')
    void withParamNameAtSign() {
        def field = Field.equal 'def', '44'
        def withParam = field.withParameterName '@unit'
        assertNotSame field, withParam, 'A new Field instance should have been created'
        assertEquals field.name, withParam.name, 'Name should have been preserved'
        assertEquals field.comparison, withParam.comparison, 'Comparison should have been preserved'
        assertEquals '@unit', withParam.parameterName, 'Parameter name not set correctly'
        assertEquals field.qualifier, withParam.qualifier, 'Qualifier should have been preserved'
    }

    @Test
    @DisplayName('withQualifier sets qualifier correctly')
    void withQualifier() {
        def field = Field.equal 'j', 'k'
        def withQual = field.withQualifier 'test'
        assertNotSame field, withQual, 'A new Field instance should have been created'
        assertEquals field.name, withQual.name, 'Name should have been preserved'
        assertEquals field.comparison, withQual.comparison, 'Comparison should have been preserved'
        assertEquals field.parameterName, withQual.parameterName, 'Parameter Name should have been preserved'
        assertEquals 'test', withQual.qualifier, 'Qualifier not set correctly'
    }

    @Test
    @DisplayName('path generates for simple unqualified PostgreSQL field')
    void pathPostgresSimpleUnqualified() {
        assertEquals("data->>'SomethingCool'",
                Field.greaterOrEqual('SomethingCool', 18).path(Dialect.POSTGRESQL, FieldFormat.SQL), 'Path not correct')
    }

    @Test
    @DisplayName('path generates for simple qualified PostgreSQL field')
    void pathPostgresSimpleQualified() {
        assertEquals("this.data->>'SomethingElse'",
                Field.less('SomethingElse', 9).withQualifier('this').path(Dialect.POSTGRESQL, FieldFormat.SQL),
                'Path not correct')
    }

    @Test
    @DisplayName('path generates for nested unqualified PostgreSQL field')
    void pathPostgresNestedUnqualified() {
        assertEquals("data#>>'{My,Nested,Field}'",
                Field.equal('My.Nested.Field', 'howdy').path(Dialect.POSTGRESQL, FieldFormat.SQL), 'Path not correct')
    }

    @Test
    @DisplayName('path generates for nested qualified PostgreSQL field')
    void pathPostgresNestedQualified() {
        assertEquals("bird.data#>>'{Nest,Away}'",
                Field.equal('Nest.Away', 'doc').withQualifier('bird').path(Dialect.POSTGRESQL, FieldFormat.SQL),
                'Path not correct')
    }

    @Test
    @DisplayName('path generates for simple unqualified SQLite field')
    void pathSQLiteSimpleUnqualified() {
        assertEquals("data->>'SomethingCool'",
                Field.greaterOrEqual('SomethingCool', 18).path(Dialect.SQLITE, FieldFormat.SQL), 'Path not correct')
    }

    @Test
    @DisplayName('path generates for simple qualified SQLite field')
    void pathSQLiteSimpleQualified() {
        assertEquals("this.data->>'SomethingElse'",
                Field.less('SomethingElse', 9).withQualifier('this').path(Dialect.SQLITE, FieldFormat.SQL),
                'Path not correct')
    }

    @Test
    @DisplayName('path generates for nested unqualified SQLite field')
    void pathSQLiteNestedUnqualified() {
        assertEquals("data->'My'->'Nested'->>'Field'",
                Field.equal('My.Nested.Field', 'howdy').path(Dialect.SQLITE, FieldFormat.SQL), 'Path not correct')
    }

    @Test
    @DisplayName('path generates for nested qualified SQLite field')
    void pathSQLiteNestedQualified() {
        assertEquals("bird.data->'Nest'->>'Away'",
                Field.equal('Nest.Away', 'doc').withQualifier('bird').path(Dialect.SQLITE, FieldFormat.SQL),
                'Path not correct')
    }

    @Test
    @DisplayName('toWhere generates for exists w/o qualifier | PostgreSQL')
    void toWhereExistsNoQualPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'that_field' IS NOT NULL", Field.exists('that_field').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for exists w/o qualifier | SQLite')
    void toWhereExistsNoQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'that_field' IS NOT NULL", Field.exists('that_field').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for not-exists w/o qualifier | PostgreSQL')
    void toWhereNotExistsNoQualPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'a_field' IS NULL", Field.notExists('a_field').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for not-exists w/o qualifier | SQLite')
    void toWhereNotExistsNoQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'a_field' IS NULL", Field.notExists('a_field').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for BETWEEN w/o qualifier, numeric range | PostgreSQL')
    void toWhereBetweenNoQualNumericPostgres() {
        ForceDialect.postgres()
        assertEquals("(data->>'age')::numeric BETWEEN @agemin AND @agemax",
                Field.between('age', 13, 17, '@age').toWhere(), 'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for BETWEEN w/o qualifier, alphanumeric range | PostgreSQL')
    void toWhereBetweenNoQualAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'city' BETWEEN :citymin AND :citymax",
                Field.between('city', 'Atlanta', 'Chicago', ':city').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for BETWEEN w/o qualifier | SQLite')
    void toWhereBetweenNoQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'age' BETWEEN @agemin AND @agemax", Field.between('age', 13, 17, '@age').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for BETWEEN w/ qualifier, numeric range | PostgreSQL')
    void toWhereBetweenQualNumericPostgres() {
        ForceDialect.postgres()
        assertEquals("(test.data->>'age')::numeric BETWEEN @agemin AND @agemax",
                Field.between('age', 13, 17, '@age').withQualifier('test').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for BETWEEN w/ qualifier, alphanumeric range | PostgreSQL')
    void toWhereBetweenQualAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals("unit.data->>'city' BETWEEN :citymin AND :citymax",
                Field.between('city', 'Atlanta', 'Chicago', ':city').withQualifier('unit').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for BETWEEN w/ qualifier | SQLite')
    void toWhereBetweenQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("my.data->>'age' BETWEEN @agemin AND @agemax",
                Field.between('age', 13, 17, '@age').withQualifier('my').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for IN/any, numeric values | PostgreSQL')
    void toWhereAnyNumericPostgres() {
        ForceDialect.postgres()
        assertEquals("(data->>'even')::numeric IN (:nbr_0, :nbr_1, :nbr_2)",
                Field.any('even', List.of(2, 4, 6), ':nbr').toWhere(), 'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for IN/any, alphanumeric values | PostgreSQL')
    void toWhereAnyAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'test' IN (:city_0, :city_1)",
                Field.any('test', List.of('Atlanta', 'Chicago'), ':city').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for IN/any | SQLite')
    void toWhereAnySQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'test' IN (:city_0, :city_1)",
                Field.any('test', List.of('Atlanta', 'Chicago'), ':city').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for inArray | PostgreSQL')
    void toWhereInArrayPostgres() {
        ForceDialect.postgres()
        assertEquals("data->'even' ??| ARRAY[:it_0, :it_1, :it_2, :it_3]",
                Field.inArray('even', 'tbl', List.of(2, 4, 6, 8), ':it').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for inArray | SQLite')
    void toWhereInArraySQLite() {
        ForceDialect.sqlite()
        assertEquals("EXISTS (SELECT 1 FROM json_each(tbl.data, '\$.test') WHERE value IN (:city_0, :city_1))",
                Field.inArray('test', 'tbl', List.of('Atlanta', 'Chicago'), ':city').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for others w/o qualifier | PostgreSQL')
    void toWhereOtherNoQualPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'some_field' = :value", Field.equal('some_field', '', ':value').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates for others w/o qualifier | SQLite')
    void toWhereOtherNoQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'some_field' = :value", Field.equal('some_field', '', ':value').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates no-parameter w/ qualifier | PostgreSQL')
    void toWhereNoParamWithQualPostgres() {
        ForceDialect.postgres()
        assertEquals("test.data->>'no_field' IS NOT NULL", Field.exists('no_field').withQualifier('test').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates no-parameter w/ qualifier | SQLite')
    void toWhereNoParamWithQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("test.data->>'no_field' IS NOT NULL", Field.exists('no_field').withQualifier('test').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates parameter w/ qualifier | PostgreSQL')
    void toWhereParamWithQualPostgres() {
        ForceDialect.postgres()
        assertEquals("(q.data->>'le_field')::numeric <= :it",
                Field.lessOrEqual('le_field', 18, ':it').withQualifier('q').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    @Test
    @DisplayName('toWhere generates parameter w/ qualifier | SQLite')
    void toWhereParamWithQualSQLite() {
        ForceDialect.sqlite()
        assertEquals("q.data->>'le_field' <= :it",
                Field.lessOrEqual('le_field', 18, ':it').withQualifier('q').toWhere(),
                'Field WHERE clause not generated correctly')
    }

    // ~~~ STATIC CONSTRUCTOR TESTS ~~~

    @Test
    @DisplayName('equal constructs a field w/o parameter name')
    void equalCtor() {
        def field = Field.equal 'Test', 14
        assertEquals 'Test', field.name, 'Field name not filled correctly'
        assertEquals Op.EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 14, field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('equal constructs a field w/ parameter name')
    void equalParameterCtor() {
        def field = Field.equal 'Test', 14, ':w'
        assertEquals 'Test', field.name, 'Field name not filled correctly'
        assertEquals Op.EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 14, field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':w', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('greater constructs a field w/o parameter name')
    void greaterCtor() {
        def field = Field.greater 'Great', 'night'
        assertEquals 'Great', field.name, 'Field name not filled correctly'
        assertEquals Op.GREATER, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'night', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('greater constructs a field w/ parameter name')
    void greaterParameterCtor() {
        def field = Field.greater 'Great', 'night', ':yeah'
        assertEquals 'Great', field.name, 'Field name not filled correctly'
        assertEquals Op.GREATER, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'night', field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':yeah', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('greaterOrEqual constructs a field w/o parameter name')
    void greaterOrEqualCtor() {
        def field = Field.greaterOrEqual 'Nice', 88L
        assertEquals 'Nice', field.name, 'Field name not filled correctly'
        assertEquals Op.GREATER_OR_EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 88L, field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('greaterOrEqual constructs a field w/ parameter name')
    void greaterOrEqualParameterCtor() {
        def field = Field.greaterOrEqual 'Nice', 88L, ':nice'
        assertEquals 'Nice', field.name, 'Field name not filled correctly'
        assertEquals Op.GREATER_OR_EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 88L, field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':nice', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('less constructs a field w/o parameter name')
    void lessCtor() {
        def field = Field.less 'Lesser', 'seven'
        assertEquals 'Lesser', field.name, 'Field name not filled correctly'
        assertEquals Op.LESS, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'seven', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('less constructs a field w/ parameter name')
    void lessParameterCtor() {
        def field = Field.less 'Lesser', 'seven', ':max'
        assertEquals 'Lesser', field.name, 'Field name not filled correctly'
        assertEquals Op.LESS, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'seven', field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':max', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('lessOrEqual constructs a field w/o parameter name')
    void lessOrEqualCtor() {
        def field = Field.lessOrEqual 'Nobody', 'KNOWS'
        assertEquals 'Nobody', field.name, 'Field name not filled correctly'
        assertEquals Op.LESS_OR_EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'KNOWS', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('lessOrEqual constructs a field w/ parameter name')
    void lessOrEqualParameterCtor() {
        def field = Field.lessOrEqual 'Nobody', 'KNOWS', ':nope'
        assertEquals 'Nobody', field.name, 'Field name not filled correctly'
        assertEquals Op.LESS_OR_EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'KNOWS', field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':nope', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('notEqual constructs a field w/o parameter name')
    void notEqualCtor() {
        def field = Field.notEqual 'Park', 'here'
        assertEquals 'Park', field.name, 'Field name not filled correctly'
        assertEquals Op.NOT_EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'here', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('notEqual constructs a field w/ parameter name')
    void notEqualParameterCtor() {
        def field = Field.notEqual 'Park', 'here', ':now'
        assertEquals 'Park', field.name, 'Field name not filled correctly'
        assertEquals Op.NOT_EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'here', field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':now', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('between constructs a field w/o parameter name')
    void betweenCtor() {
        def field = Field.between 'Age', 18, 49
        assertEquals 'Age', field.name, 'Field name not filled correctly'
        assertEquals Op.BETWEEN, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 18, field.comparison.value.first, 'Field comparison min value not filled correctly'
        assertEquals 49, field.comparison.value.second, 'Field comparison max value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('between constructs a field w/ parameter name')
    void betweenParameterCtor() {
        def field = Field.between 'Age', 18, 49, ':limit'
        assertEquals 'Age', field.name, 'Field name not filled correctly'
        assertEquals Op.BETWEEN, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 18, field.comparison.value.first, 'Field comparison min value not filled correctly'
        assertEquals 49, field.comparison.value.second, 'Field comparison max value not filled correctly'
        assertEquals ':limit', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('any constructs a field w/o parameter name')
    void anyCtor() {
        def field = Field.any 'Here', List.of(8, 16, 32)
        assertEquals 'Here', field.name, 'Field name not filled correctly'
        assertEquals Op.IN, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals List.of(8, 16, 32), field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('any constructs a field w/ parameter name')
    void anyParameterCtor() {
        def field = Field.any 'Here', List.of(8, 16, 32), ':list'
        assertEquals 'Here', field.name, 'Field name not filled correctly'
        assertEquals Op.IN, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals List.of(8, 16, 32), field.comparison.value, 'Field comparison value not filled correctly'
        assertEquals ':list', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('inArray constructs a field w/o parameter name')
    void inArrayCtor() {
        def field = Field.inArray 'ArrayField', 'table', List.of('z')
        assertEquals 'ArrayField', field.name, 'Field name not filled correctly'
        assertEquals Op.IN_ARRAY, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'table', field.comparison.value.first, 'Field comparison table not filled correctly'
        assertEquals List.of('z'), field.comparison.value.second, 'Field comparison values not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('inArray constructs a field w/ parameter name')
    void inArrayParameterCtor() {
        def field = Field.inArray 'ArrayField', 'table', List.of('z'), ':a'
        assertEquals 'ArrayField', field.name, 'Field name not filled correctly'
        assertEquals Op.IN_ARRAY, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals 'table', field.comparison.value.first, 'Field comparison table not filled correctly'
        assertEquals List.of('z'), field.comparison.value.second, 'Field comparison values not filled correctly'
        assertEquals ':a', field.parameterName, 'Field parameter name not filled correctly'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('exists constructs a field')
    void existsCtor() {
        def field = Field.exists 'Groovy'
        assertEquals 'Groovy', field.name, 'Field name not filled correctly'
        assertEquals Op.EXISTS, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals '', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('notExists constructs a field')
    void notExistsCtor() {
        def field = Field.notExists 'Groovy'
        assertEquals 'Groovy', field.name, 'Field name not filled correctly'
        assertEquals Op.NOT_EXISTS, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals '', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('named constructs a field')
    void namedCtor() {
        def field = Field.named 'Tacos'
        assertEquals 'Tacos', field.name, 'Field name not filled correctly'
        assertEquals Op.EQUAL, field.comparison.op, 'Field comparison operation not filled correctly'
        assertEquals '', field.comparison.value, 'Field comparison value not filled correctly'
        assertNull field.parameterName, 'The parameter name should have been null'
        assertNull field.qualifier, 'The qualifier should have been null'
    }

    @Test
    @DisplayName('static constructors fail for invalid parameter name')
    void staticCtorsFailOnParamName() {
        assertThrows(DocumentException) { Field.equal('a', 'b', "that ain't it, Jack...") }
    }

    @Test
    @DisplayName('nameToPath creates a simple PostgreSQL SQL name')
    void nameToPathPostgresSimpleSQL() {
        assertEquals("data->>'Simple'", Field.nameToPath('Simple', Dialect.POSTGRESQL, FieldFormat.SQL),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a simple SQLite SQL name')
    void nameToPathSQLiteSimpleSQL() {
        assertEquals("data->>'Simple'", Field.nameToPath('Simple', Dialect.SQLITE, FieldFormat.SQL),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a nested PostgreSQL SQL name')
    void nameToPathPostgresNestedSQL() {
        assertEquals("data#>>'{A,Long,Path,to,the,Property}'",
                Field.nameToPath('A.Long.Path.to.the.Property', Dialect.POSTGRESQL, FieldFormat.SQL),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a nested SQLite SQL name')
    void nameToPathSQLiteNestedSQL() {
        assertEquals("data->'A'->'Long'->'Path'->'to'->'the'->>'Property'",
                Field.nameToPath('A.Long.Path.to.the.Property', Dialect.SQLITE, FieldFormat.SQL),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a simple PostgreSQL JSON name')
    void nameToPathPostgresSimpleJSON() {
        assertEquals("data->'Simple'", Field.nameToPath('Simple', Dialect.POSTGRESQL, FieldFormat.JSON),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a simple SQLite JSON name')
    void nameToPathSQLiteSimpleJSON() {
        assertEquals("data->'Simple'", Field.nameToPath('Simple', Dialect.SQLITE, FieldFormat.JSON),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a nested PostgreSQL JSON name')
    void nameToPathPostgresNestedJSON() {
        assertEquals("data#>'{A,Long,Path,to,the,Property}'",
                Field.nameToPath('A.Long.Path.to.the.Property', Dialect.POSTGRESQL, FieldFormat.JSON),
                'Path not constructed correctly')
    }

    @Test
    @DisplayName('nameToPath creates a nested SQLite JSON name')
    void nameToPathSQLiteNestedJSON() {
        assertEquals("data->'A'->'Long'->'Path'->'to'->'the'->'Property'",
                Field.nameToPath('A.Long.Path.to.the.Property', Dialect.SQLITE, FieldFormat.JSON),
                'Path not constructed correctly')
    }
}
