package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import solutions.bitbadger.documents.query.QueryUtils

import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `QueryUtils` class
 */
@DisplayName('Groovy | Query | QueryUtils')
class QueryUtilsTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    void cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName('statementWhere generates correctly')
    void statementWhere() {
        assertEquals 'x WHERE y', QueryUtils.statementWhere('x', 'y'), 'Statements not combined correctly'
    }

    @Test
    @DisplayName('byId generates a numeric ID query | PostgreSQL')
    void byIdNumericPostgres() {
        ForceDialect.postgres()
        assertEquals "test WHERE (data->>'id')::numeric = :id", QueryUtils.byId('test', 9)
    }

    @Test
    @DisplayName('byId generates an alphanumeric ID query | PostgreSQL')
    void byIdAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals "unit WHERE data->>'id' = :id", QueryUtils.byId('unit', '18')
    }

    @Test
    @DisplayName('byId generates ID query | SQLite')
    void byIdSQLite() {
        ForceDialect.sqlite()
        assertEquals "yo WHERE data->>'id' = :id", QueryUtils.byId('yo', 27)
    }

    @Test
    @DisplayName('byFields generates default field query | PostgreSQL')
    void byFieldsMultipleDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals("this WHERE data->>'a' = :the_a AND (data->>'b')::numeric = :b_value",
                QueryUtils.byFields('this', List.of(Field.equal('a', '', ':the_a'), Field.equal('b', 0, ':b_value'))))
    }

    @Test
    @DisplayName('byFields generates default field query | SQLite')
    void byFieldsMultipleDefaultSQLite() {
        ForceDialect.sqlite()
        assertEquals("this WHERE data->>'a' = :the_a AND data->>'b' = :b_value",
                QueryUtils.byFields('this', List.of(Field.equal('a', '', ':the_a'), Field.equal('b', 0, ':b_value'))))
    }

    @Test
    @DisplayName('byFields generates ANY field query | PostgreSQL')
    void byFieldsMultipleAnyPostgres() {
        ForceDialect.postgres()
        assertEquals("that WHERE data->>'a' = :the_a OR (data->>'b')::numeric = :b_value",
                QueryUtils.byFields('that', List.of(Field.equal('a', '', ':the_a'), Field.equal('b', 0, ':b_value')),
                        FieldMatch.ANY))
    }

    @Test
    @DisplayName('byFields generates ANY field query | SQLite')
    void byFieldsMultipleAnySQLite() {
        ForceDialect.sqlite()
        assertEquals("that WHERE data->>'a' = :the_a OR data->>'b' = :b_value",
                QueryUtils.byFields('that', List.of(Field.equal('a', '', ':the_a'), Field.equal('b', 0, ':b_value')),
                        FieldMatch.ANY))
    }

    @Test
    @DisplayName('orderBy generates for no fields')
    void orderByNone() {
        assertEquals('', QueryUtils.orderBy(List.of(), Dialect.POSTGRESQL),
                'ORDER BY should have been blank (PostgreSQL)')
        assertEquals '', QueryUtils.orderBy(List.of(), Dialect.SQLITE), 'ORDER BY should have been blank (SQLite)'
    }

    @Test
    @DisplayName('orderBy generates single, no direction | PostgreSQL')
    void orderBySinglePostgres() {
        assertEquals(" ORDER BY data->>'TestField'",
                QueryUtils.orderBy(List.of(Field.named('TestField')), Dialect.POSTGRESQL),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates single, no direction | SQLite')
    void orderBySingleSQLite() {
        assertEquals(" ORDER BY data->>'TestField'",
                QueryUtils.orderBy(List.of(Field.named('TestField')), Dialect.SQLITE),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates multiple with direction | PostgreSQL')
    void orderByMultiplePostgres() {
        assertEquals(" ORDER BY data#>>'{Nested,Test,Field}' DESC, data->>'AnotherField', data->>'It' DESC",
                QueryUtils.orderBy(List.of(Field.named('Nested.Test.Field DESC'), Field.named('AnotherField'),
                                Field.named('It DESC')),
                        Dialect.POSTGRESQL),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates multiple with direction | SQLite')
    void orderByMultipleSQLite() {
        assertEquals(" ORDER BY data->'Nested'->'Test'->>'Field' DESC, data->>'AnotherField', data->>'It' DESC",
                QueryUtils.orderBy(List.of(Field.named('Nested.Test.Field DESC'), Field.named('AnotherField'),
                                Field.named('It DESC')),
                        Dialect.SQLITE),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates numeric ordering | PostgreSQL')
    void orderByNumericPostgres() {
        assertEquals(" ORDER BY (data->>'Test')::numeric",
                QueryUtils.orderBy(List.of(Field.named('n:Test')), Dialect.POSTGRESQL),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates numeric ordering | SQLite')
    void orderByNumericSQLite() {
        assertEquals(" ORDER BY data->>'Test'", QueryUtils.orderBy(List.of(Field.named('n:Test')), Dialect.SQLITE),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates case-insensitive ordering | PostgreSQL')
    void orderByCIPostgres() {
        assertEquals(" ORDER BY LOWER(data#>>'{Test,Field}') DESC NULLS FIRST",
                QueryUtils.orderBy(List.of(Field.named('i:Test.Field DESC NULLS FIRST')), Dialect.POSTGRESQL),
                'ORDER BY not constructed correctly')
    }

    @Test
    @DisplayName('orderBy generates case-insensitive ordering | SQLite')
    void orderByCISQLite() {
        assertEquals(" ORDER BY data->'Test'->>'Field' COLLATE NOCASE ASC NULLS LAST",
                QueryUtils.orderBy(List.of(Field.named('i:Test.Field ASC NULLS LAST')), Dialect.SQLITE),
                'ORDER BY not constructed correctly')
    }
}
