package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import solutions.bitbadger.documents.query.Where

import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `Where` object
 */
@DisplayName('Groovy | Query | Where')
class WhereTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    void cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName('byFields is blank when given no fields')
    void byFieldsBlankIfEmpty() {
        assertEquals '', Where.byFields(List.of())
    }

    @Test
    @DisplayName('byFields generates one numeric field | PostgreSQL')
    void byFieldsOneFieldPostgres() {
        ForceDialect.postgres()
        assertEquals "(data->>'it')::numeric = :that", Where.byFields(List.of(Field.equal('it', 9, ':that')))
    }

    @Test
    @DisplayName('byFields generates one alphanumeric field | PostgreSQL')
    void byFieldsOneAlphaFieldPostgres() {
        ForceDialect.postgres()
        assertEquals "data->>'it' = :that", Where.byFields(List.of(Field.equal('it', '', ':that')))
    }

    @Test
    @DisplayName('byFields generates one field | SQLite')
    void byFieldsOneFieldSQLite() {
        ForceDialect.sqlite()
        assertEquals "data->>'it' = :that", Where.byFields(List.of(Field.equal('it', '', ':that')))
    }

    @Test
    @DisplayName('byFields generates multiple fields w/ default match | PostgreSQL')
    void byFieldsMultipleDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'1' = :one AND (data->>'2')::numeric = :two AND data->>'3' = :three",
                Where.byFields(
                        List.of(Field.equal('1', '', ':one'), Field.equal('2', 0L, ':two'),
                                Field.equal('3', '', ':three'))))
    }

    @Test
    @DisplayName('byFields generates multiple fields w/ default match | SQLite')
    void byFieldsMultipleDefaultSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'1' = :one AND data->>'2' = :two AND data->>'3' = :three",
                Where.byFields(
                        List.of(Field.equal('1', '', ':one'), Field.equal('2', 0L, ':two'),
                                Field.equal('3', '', ':three'))))
    }

    @Test
    @DisplayName('byFields generates multiple fields w/ ANY match | PostgreSQL')
    void byFieldsMultipleAnyPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'1' = :one OR (data->>'2')::numeric = :two OR data->>'3' = :three",
                Where.byFields(
                        List.of(Field.equal('1', '', ':one'), Field.equal('2', 0L, ':two'),
                                Field.equal('3', '', ':three')),
                        FieldMatch.ANY))
    }

    @Test
    @DisplayName('byFields generates multiple fields w/ ANY match | SQLite')
    void byFieldsMultipleAnySQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'1' = :one OR data->>'2' = :two OR data->>'3' = :three",
                Where.byFields(
                        List.of(Field.equal('1', '', ':one'), Field.equal('2', 0L, ':two'),
                                Field.equal('3', '', ':three')),
                        FieldMatch.ANY))
    }

    @Test
    @DisplayName('byId generates defaults for alphanumeric key | PostgreSQL')
    void byIdDefaultAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals "data->>'id' = :id", Where.byId()
    }

    @Test
    @DisplayName('byId generates defaults for numeric key | PostgreSQL')
    void byIdDefaultNumericPostgres() {
        ForceDialect.postgres()
        assertEquals "(data->>'id')::numeric = :id", Where.byId(":id", 5)
    }

    @Test
    @DisplayName('byId generates defaults | SQLite')
    void byIdDefaultSQLite() {
        ForceDialect.sqlite()
        assertEquals "data->>'id' = :id", Where.byId()
    }

    @Test
    @DisplayName('byId generates named ID | PostgreSQL')
    void byIdDefaultNamedPostgres() {
        ForceDialect.postgres()
        assertEquals "data->>'id' = :key", Where.byId(':key')
    }

    @Test
    @DisplayName('byId generates named ID | SQLite')
    void byIdDefaultNamedSQLite() {
        ForceDialect.sqlite()
        assertEquals "data->>'id' = :key", Where.byId(':key')
    }

    @Test
    @DisplayName('jsonContains generates defaults | PostgreSQL')
    void jsonContainsDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals 'data @> :criteria', Where.jsonContains()
    }

    @Test
    @DisplayName('jsonContains generates named parameter | PostgreSQL')
    void jsonContainsNamedPostgres() {
        ForceDialect.postgres()
        assertEquals 'data @> :it', Where.jsonContains(':it')
    }

    @Test
    @DisplayName('jsonContains fails | SQLite')
    void jsonContainsFailsSQLite() {
        ForceDialect.sqlite()
        assertThrows(DocumentException) { Where.jsonContains() }
    }

    @Test
    @DisplayName('jsonPathMatches generates defaults | PostgreSQL')
    void jsonPathMatchDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals 'jsonb_path_exists(data, :path::jsonpath)', Where.jsonPathMatches()
    }

    @Test
    @DisplayName('jsonPathMatches generates named parameter | PostgreSQL')
    void jsonPathMatchNamedPostgres() {
        ForceDialect.postgres()
        assertEquals 'jsonb_path_exists(data, :jp::jsonpath)', Where.jsonPathMatches(':jp')
    }

    @Test
    @DisplayName('jsonPathMatches fails | SQLite')
    void jsonPathFailsSQLite() {
        ForceDialect.sqlite()
        assertThrows(DocumentException) { Where.jsonPathMatches() }
    }
}
