package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

final class FindFunctions {

    private static String docIds(List<JsonDocument> docs) {
        return docs*.id.inject('') { acc, it -> acc == '' ? it : "$acc|$it" }
    }

    static void allDefault(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals 5, db.conn.findAll(TEST_TABLE, JsonDocument).size(), 'There should have been 5 documents returned'
    }

    static void allAscending(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findAll TEST_TABLE, JsonDocument, List.of(Field.named('id'))
        assertEquals 5, docs.size(), 'There should have been 5 documents returned'
        assertEquals 'five|four|one|three|two', docIds(docs), 'The documents were not ordered correctly'
    }

    static void allDescending(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findAll TEST_TABLE, JsonDocument, List.of(Field.named('id DESC'))
        assertEquals 5, docs.size(), 'There should have been 5 documents returned'
        assertEquals 'two|three|one|four|five', docIds(docs), 'The documents were not ordered correctly'
    }

    static void allNumOrder(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findAll(TEST_TABLE, JsonDocument,
                List.of(Field.named('sub.foo NULLS LAST'), Field.named('n:numValue')))
        assertEquals 5, docs.size(), 'There should have been 5 documents returned'
        assertEquals 'two|four|one|three|five', docIds(docs), 'The documents were not ordered correctly'
    }

    static void allEmpty(ThrowawayDatabase db) {
        assertEquals 0, db.conn.findAll(TEST_TABLE, JsonDocument).size(), 'There should have been no documents returned'
    }

    static void byIdString(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findById TEST_TABLE, 'two', JsonDocument
        assertTrue doc.isPresent(), 'The document should have been returned'
        assertEquals 'two', doc.get().id, 'An incorrect document was returned'
    }

    static void byIdNumber(ThrowawayDatabase db) {
        Configuration.idField = 'key'
        try {
            db.conn.insert TEST_TABLE, new NumIdDocument(18, 'howdy')
            assertTrue(db.conn.findById(TEST_TABLE, 18, NumIdDocument).isPresent(),
                    'The document should have been returned')
        } finally {
            Configuration.idField = 'id'
        }
    }

    static void byIdNotFound(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse(db.conn.findById(TEST_TABLE, 'x', JsonDocument).isPresent(),
                'There should have been no document returned')
    }

    static void byFieldsMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByFields(TEST_TABLE,
                List.of(Field.any('value', List.of('blue', 'purple')), Field.exists('sub')), JsonDocument,
                FieldMatch.ALL)
        assertEquals 1, docs.size(), 'There should have been a document returned'
        assertEquals 'four', docs[0].id, 'The incorrect document was returned'
    }

    static void byFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByFields(TEST_TABLE, List.of(Field.equal('value', 'purple')), JsonDocument, null,
                List.of(Field.named('id')))
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        assertEquals 'five|four', docIds(docs), 'The documents were not ordered correctly'
    }

    static void byFieldsMatchNumIn(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByFields TEST_TABLE, List.of(Field.any('numValue', List.of(2, 4, 6, 8))), JsonDocument
        assertEquals 1, docs.size(), 'There should have been a document returned'
        assertEquals 'three', docs[0].id, 'The incorrect document was returned'
    }

    static void byFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals(0, db.conn.findByFields(TEST_TABLE, List.of(Field.greater('numValue', 100)), JsonDocument).size(),
                'There should have been no documents returned')
    }

    static void byFieldsMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert TEST_TABLE, it }
        def docs = db.conn.findByFields(TEST_TABLE, List.of(Field.inArray('values', TEST_TABLE, List.of('c'))),
                        ArrayDocument)
        assertEquals 2, docs.size(), 'There should have been two documents returned'
        assertTrue(List.of('first', 'second').contains(docs[0].id),
                "An incorrect document was returned (${docs[0].id})")
        assertTrue(List.of('first', 'second').contains(docs[1].id),
                "An incorrect document was returned (${docs[1].id})")
    }

    static void byFieldsNoMatchInArray(ThrowawayDatabase db) {
        ArrayDocument.testDocuments.forEach { db.conn.insert TEST_TABLE, it }
        assertEquals(0,
                db.conn.findByFields(TEST_TABLE, List.of(Field.inArray('values', TEST_TABLE, List.of('j'))),
                        ArrayDocument).size(),
                'There should have been no documents returned')
    }

    static void byContainsMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByContains TEST_TABLE, Map.of('value', 'purple'), JsonDocument
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        assertTrue List.of('four', 'five').contains(docs[0].id), "An incorrect document was returned (${docs[0].id})"
        assertTrue List.of('four', 'five').contains(docs[1].id), "An incorrect document was returned (${docs[1].id})"
    }

    static void byContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByContains(TEST_TABLE, Map.of('sub', Map.of('foo', 'green')), JsonDocument,
                List.of(Field.named('value')))
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        assertEquals 'two|four', docIds(docs), 'The documents were not ordered correctly'
    }

    static void byContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals(0, db.conn.findByContains(TEST_TABLE, Map.of('value', 'indigo'), JsonDocument).size(),
                'There should have been no documents returned')
    }

    static void byJsonPathMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByJsonPath TEST_TABLE, '$.numValue ? (@ > 10)', JsonDocument
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        assertTrue List.of('four', 'five').contains(docs[0].id), "An incorrect document was returned (${docs[0].id})"
        assertTrue List.of('four', 'five').contains(docs[1].id), "An incorrect document was returned (${docs[1].id})"
    }

    static void byJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load db
        def docs = db.conn.findByJsonPath TEST_TABLE, '$.numValue ? (@ > 10)', JsonDocument, List.of(Field.named('id'))
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        assertEquals 'five|four', docIds(docs), 'The documents were not ordered correctly'
    }

    static void byJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertEquals(0, db.conn.findByJsonPath(TEST_TABLE, '$.numValue ? (@ > 100)', JsonDocument).size(),
                'There should have been no documents returned')
    }

    static void firstByFieldsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByFields TEST_TABLE, List.of(Field.equal('value', 'another')), JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'two', doc.get().id, 'The incorrect document was returned'
    }

    static void firstByFieldsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByFields TEST_TABLE, List.of(Field.equal('sub.foo', 'green')), JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertTrue List.of('two', 'four').contains(doc.get().id), "An incorrect document was returned (${doc.get().id})"
    }

    static void firstByFieldsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByFields(TEST_TABLE, List.of(Field.equal('sub.foo', 'green')), JsonDocument, null,
                List.of(Field.named('n:numValue DESC')))
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'four', doc.get().id, 'An incorrect document was returned'
    }

    static void firstByFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse(db.conn.findFirstByFields(TEST_TABLE, List.of(Field.equal('value', 'absent')), JsonDocument)
                        .isPresent(),
                'There should have been no document returned')
    }

    static void firstByContainsMatchOne(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByContains TEST_TABLE, Map.of('value', 'FIRST!'), JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'one', doc.get().id, 'An incorrect document was returned'
    }

    static void firstByContainsMatchMany(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByContains TEST_TABLE, Map.of('value', 'purple'), JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertTrue(List.of('four', 'five').contains(doc.get().id),
                "An incorrect document was returned (${doc.get().id})")
    }

    static void firstByContainsMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByContains(TEST_TABLE, Map.of('value', 'purple'), JsonDocument,
                List.of(Field.named('sub.bar NULLS FIRST')))
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'five', doc.get().id, 'An incorrect document was returned'
    }

    static void firstByContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse(db.conn.findFirstByContains(TEST_TABLE, Map.of('value', 'indigo'), JsonDocument).isPresent(),
                'There should have been no document returned')
    }

    static void firstByJsonPathMatchOne(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByJsonPath TEST_TABLE, '$.numValue ? (@ == 10)', JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'two', doc.get().id, 'An incorrect document was returned'
    }

    static void firstByJsonPathMatchMany(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByJsonPath TEST_TABLE, '$.numValue ? (@ > 10)', JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertTrue(List.of('four', 'five').contains(doc.get().id),
                "An incorrect document was returned (${doc.get().id})")
    }

    static void firstByJsonPathMatchOrdered(ThrowawayDatabase db) {
        JsonDocument.load db
        def doc = db.conn.findFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 10)', JsonDocument,
                List.of(Field.named('id DESC')))
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals 'four', doc.get().id, 'An incorrect document was returned'
    }

    static void firstByJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse(db.conn.findFirstByJsonPath(TEST_TABLE, '$.numValue ? (@ > 100)', JsonDocument).isPresent(),
                'There should have been no document returned')
    }
}
