package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.Field

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

final class RemoveFieldsFunctions {

    static void byIdMatchFields(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.removeFieldsById TEST_TABLE, 'two', List.of('sub', 'value')
        def doc = db.conn.findById TEST_TABLE, 'two', JsonDocument
        assertTrue doc.isPresent(), 'There should have been a document returned'
        assertEquals '', doc.get().value, 'The value should have been empty'
        assertNull doc.get().sub, 'The sub-document should have been removed'
    }

    static void byIdMatchNoFields(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsByFields(TEST_TABLE, List.of(Field.exists('a_field_that_does_not_exist')))
        db.conn.removeFieldsById TEST_TABLE, 'one', List.of('a_field_that_does_not_exist') // no exception = pass
    }

    static void byIdNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsById(TEST_TABLE, 'fifty')
        db.conn.removeFieldsById TEST_TABLE, 'fifty', List.of('sub') // no exception = pass
    }

    static void byFieldsMatchFields(ThrowawayDatabase db) {
        JsonDocument.load db
        def fields = List.of Field.equal('numValue', 17)
        db.conn.removeFieldsByFields TEST_TABLE, fields, List.of('sub')
        def doc = db.conn.findFirstByFields TEST_TABLE, fields, JsonDocument
        assertTrue doc.isPresent(), 'The document should have been returned'
        assertEquals 'four', doc.get().id, 'An incorrect document was returned'
        assertNull doc.get().sub, 'The sub-document should have been removed'
    }

    static void byFieldsMatchNoFields(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsByFields(TEST_TABLE, List.of(Field.exists('nada')))
        db.conn.removeFieldsByFields TEST_TABLE, List.of(Field.equal('numValue', 17)), List.of('nada') // no exn = pass
    }

    static void byFieldsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def fields = List.of Field.notEqual('missing', 'nope')
        assertFalse db.conn.existsByFields(TEST_TABLE, fields)
        db.conn.removeFieldsByFields TEST_TABLE, fields, List.of('value') // no exception = pass
    }

    static void byContainsMatchFields(ThrowawayDatabase db) {
        JsonDocument.load db
        def criteria = Map.of('sub', Map.of('foo', 'green'))
        db.conn.removeFieldsByContains TEST_TABLE, criteria, List.of('value')
        def docs = db.conn.findByContains TEST_TABLE, criteria, JsonDocument
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        docs.forEach {
            assertTrue List.of('two', 'four').contains(it.id), "An incorrect document was returned (${it.id})"
            assertEquals '', it.value, 'The value should have been empty'
        }
    }

    static void byContainsMatchNoFields(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsByFields(TEST_TABLE, List.of(Field.exists('invalid_field')))
        db.conn.removeFieldsByContains TEST_TABLE, Map.of('sub', Map.of('foo', 'green')), List.of('invalid_field')
        // no exception = pass
    }

    static void byContainsNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def contains = Map.of 'value', 'substantial'
        assertFalse db.conn.existsByContains(TEST_TABLE, contains)
        db.conn.removeFieldsByContains TEST_TABLE, contains, List.of('numValue')
    }

    static void byJsonPathMatchFields(ThrowawayDatabase db) {
        JsonDocument.load db
        def path = '$.value ? (@ == "purple")'
        db.conn.removeFieldsByJsonPath TEST_TABLE, path, List.of('sub')
        def docs = db.conn.findByJsonPath TEST_TABLE, path, JsonDocument
        assertEquals 2, docs.size(), 'There should have been 2 documents returned'
        docs.forEach {
            assertTrue List.of('four', 'five').contains(it.id), "An incorrect document was returned (${it.id})"
            assertNull it.sub, 'The sub-document should have been removed'
        }
    }

    static void byJsonPathMatchNoFields(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsByFields(TEST_TABLE, List.of(Field.exists('submarine')))
        db.conn.removeFieldsByJsonPath TEST_TABLE, '$.value ? (@ == "purple")', List.of('submarine') // no exn = pass
    }

    static void byJsonPathNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        def path = '$.value ? (@ == "mauve")'
        assertFalse db.conn.existsByJsonPath(TEST_TABLE, path)
        db.conn.removeFieldsByJsonPath TEST_TABLE, path, List.of('value') // no exception = pass
    }
}
