package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.java.RemoveFields as JvmRemoveFields
import solutions.bitbadger.documents.kotlinx.extensions.*
import solutions.bitbadger.documents.query.RemoveFieldsQuery
import java.sql.Connection

/**
 * Functions to remove fields from documents
 */
object RemoveFields {

    /**
     * Remove fields from a document by its ID
     *
     * @param tableName The name of the table in which the document's fields should be removed
     * @param docId The ID of the document to have fields removed
     * @param toRemove The names of the fields to be removed
     * @param conn The connection on which the update should be executed
     */
    fun <TKey> byId(tableName: String, docId: TKey, toRemove: Collection<String>, conn: Connection) =
        JvmRemoveFields.byId(tableName, docId, toRemove, conn)

    /**
     * Remove fields from a document by its ID
     *
     * @param tableName The name of the table in which the document's fields should be removed
     * @param docId The ID of the document to have fields removed
     * @param toRemove The names of the fields to be removed
     */
    fun <TKey> byId(tableName: String, docId: TKey, toRemove: Collection<String>) =
        JvmRemoveFields.byId(tableName, docId, toRemove)

    /**
     * Remove fields from documents using a field comparison
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param fields The fields which should be compared
     * @param toRemove The names of the fields to be removed
     * @param howMatched How the fields should be matched
     * @param conn The connection on which the update should be executed
     */
    fun byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        toRemove: Collection<String>,
        howMatched: FieldMatch? = null,
        conn: Connection
    ) =
        JvmRemoveFields.byFields(tableName, fields, toRemove, howMatched, conn)

    /**
     * Remove fields from documents using a field comparison
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param fields The fields which should be compared
     * @param toRemove The names of the fields to be removed
     * @param howMatched How the fields should be matched
     */
    fun byFields(
        tableName: String,
        fields: Collection<Field<*>>,
        toRemove: Collection<String>,
        howMatched: FieldMatch? = null
    ) =
        JvmRemoveFields.byFields(tableName, fields, toRemove, howMatched)

    /**
     * Remove fields from documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param criteria The object against which JSON containment should be checked
     * @param toRemove The names of the fields to be removed
     * @param conn The connection on which the update should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(
        tableName: String,
        criteria: TContains,
        toRemove: Collection<String>,
        conn: Connection
    ) {
        val nameParams = Parameters.fieldNames(toRemove)
        conn.customNonQuery(
            RemoveFieldsQuery.byContains(tableName, nameParams),
            listOf(Parameters.json(":criteria", criteria), *nameParams.toTypedArray())
        )
    }

    /**
     * Remove fields from documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param criteria The object against which JSON containment should be checked
     * @param toRemove The names of the fields to be removed
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(tableName: String, criteria: TContains, toRemove: Collection<String>) =
        Configuration.dbConn().use { byContains(tableName, criteria, toRemove, it) }

    /**
     * Remove fields from documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param path The JSON path comparison to match
     * @param toRemove The names of the fields to be removed
     * @param conn The connection on which the update should be executed
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String, toRemove: Collection<String>, conn: Connection) =
        JvmRemoveFields.byJsonPath(tableName, path, toRemove, conn)

    /**
     * Remove fields from documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param path The JSON path comparison to match
     * @param toRemove The names of the fields to be removed
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String, toRemove: Collection<String>) =
        JvmRemoveFields.byJsonPath(tableName, path, toRemove)
}
