package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.java.Results as CoreResults
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.sql.SQLException

/**
 * Helper functions for handling results
 */
object Results {

    /**
     * Create a domain item from a document, specifying the field in which the document is found
     *
     * @param field The field name containing the JSON document
     * @return A function to create the constructed  domain item
     */
    inline fun <reified TDoc> fromDocument(field: String): (ResultSet) -> TDoc =
        { rs -> DocumentConfig.deserialize<TDoc>(rs.getString(field)) }

    /**
     * Create a domain item from a document
     *
     * @return The constructed domain item
     */
    inline fun <reified TDoc> fromData(rs: ResultSet) =
        fromDocument<TDoc>("data")(rs)

    /**
     * Create a list of items for the results of the given command, using the specified mapping function
     *
     * @param stmt The prepared statement to execute
     * @param mapFunc The mapping function from data reader to domain class instance
     * @return A list of items from the query's result
     * @throws DocumentException If there is a problem executing the query
     */
    inline fun <reified TDoc : Any> toCustomList(stmt: PreparedStatement, mapFunc: (ResultSet) -> TDoc) =
        try {
            stmt.executeQuery().use {
                val results = mutableListOf<TDoc>()
                while (it.next()) {
                    results.add(mapFunc(it))
                }
                results.toList()
            }
        } catch (ex: SQLException) {
            throw DocumentException("Error retrieving documents from query: ${ex.message}", ex)
        }

    /**
     * Extract a count from the first column
     *
     * @param rs A `ResultSet` set to the row with the count to retrieve
     * @return The count from the row
     */
    fun toCount(rs: ResultSet) =
        when (Configuration.dialect()) {
            Dialect.POSTGRESQL -> rs.getInt("it").toLong()
            Dialect.SQLITE -> rs.getLong("it")
        }

    /**
     * Extract a true/false value from the first column
     *
     * @param rs A `ResultSet` set to the row with the true/false value to retrieve
     * @return The true/false value from the row
     */
    fun toExists(rs: ResultSet) =
        when (Configuration.dialect()) {
            Dialect.POSTGRESQL -> rs.getBoolean("it")
            Dialect.SQLITE -> toCount(rs) > 0L
        }

    /**
     * Retrieve the JSON text of a document, specifying the field in which the document is found
     *
     * @param field The field name containing the JSON document
     * @param rs A `ResultSet` set to the row with the document to be constructed
     * @return The JSON text of the document
     */
    fun jsonFromDocument(field: String, rs: ResultSet) =
        CoreResults.jsonFromDocument(field, rs)

    /**
     * Retrieve the JSON text of a document, specifying the field in which the document is found
     *
     * @param rs A `ResultSet` set to the row with the document to be constructed
     * @return The JSON text of the document
     */
    fun jsonFromData(rs: ResultSet) =
        CoreResults.jsonFromData(rs)

    /**
     * Create a JSON array of items for the results of the given command, using the specified mapping function
     *
     * @param stmt The prepared statement to execute
     * @param mapFunc The mapping function from data reader to JSON text
     * @return A string with a JSON array of documents from the query's result
     * @throws DocumentException If there is a problem executing the query (unchecked)
     */
    fun toJsonArray(stmt: PreparedStatement, mapFunc: (ResultSet) -> String) =
        CoreResults.toJsonArray(stmt, mapFunc)
}
