package solutions.bitbadger.documents.kotlinx.tests.integration

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.kotlinx.Results
import solutions.bitbadger.documents.kotlinx.extensions.*
import solutions.bitbadger.documents.kotlinx.tests.ArrayDocument
import solutions.bitbadger.documents.kotlinx.tests.JsonDocument
import solutions.bitbadger.documents.kotlinx.tests.TEST_TABLE
import solutions.bitbadger.documents.query.*
import java.io.PrintWriter
import java.io.StringWriter
import kotlin.test.assertEquals
import kotlin.test.assertNotNull
import kotlin.test.assertNull

/**
 * Integration tests for the `Custom` object
 */
object CustomFunctions {

    fun listEmpty(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        db.conn.deleteByFields(TEST_TABLE, listOf(Field.exists(Configuration.idField)))
        val result = db.conn.customList<JsonDocument>(FindQuery.all(TEST_TABLE), mapFunc = Results::fromData)
        assertEquals(0, result.size, "There should have been no results")
    }

    fun listAll(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val result = db.conn.customList<JsonDocument>(FindQuery.all(TEST_TABLE), mapFunc = Results::fromData)
        assertEquals(5, result.size, "There should have been 5 results")
    }

    fun jsonArrayEmpty(db: ThrowawayDatabase) {
        assertEquals(0L, db.conn.countAll(TEST_TABLE), "The test table should be empty")
        assertEquals(
            "[]",
            db.conn.customJsonArray(FindQuery.all(TEST_TABLE), listOf(), Results::jsonFromData),
            "An empty list was not represented correctly"
        )
    }

    fun jsonArraySingle(db: ThrowawayDatabase) {
        db.conn.insert(TEST_TABLE, ArrayDocument("one", listOf("2", "3")))
        assertEquals(
            JsonFunctions.maybeJsonB("""[{"id":"one","values":["2","3"]}]"""),
            db.conn.customJsonArray(FindQuery.all(TEST_TABLE), listOf(), Results::jsonFromData),
            "A single document list was not represented correctly"
        )
    }

    fun jsonArrayMany(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        assertEquals(
            JsonFunctions.maybeJsonB("""[{"id":"first","values":["a","b","c"]},"""
                    + """{"id":"second","values":["c","d","e"]},{"id":"third","values":["x","y","z"]}]"""),
            db.conn.customJsonArray(FindQuery.all(TEST_TABLE) + orderBy(listOf(Field.named("id"))), listOf(),
                Results::jsonFromData),
            "A multiple document list was not represented correctly"
        )
    }

    fun writeJsonArrayEmpty(db: ThrowawayDatabase) {
        assertEquals(0L, db.conn.countAll(TEST_TABLE), "The test table should be empty")
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE), listOf(), writer, Results::jsonFromData)
        assertEquals("[]", output.toString(), "An empty list was not represented correctly")
    }

    fun writeJsonArraySingle(db: ThrowawayDatabase) {
        db.conn.insert(TEST_TABLE, ArrayDocument("one", listOf("2", "3")))
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE), listOf(), writer, Results::jsonFromData)
        assertEquals(
            JsonFunctions.maybeJsonB("""[{"id":"one","values":["2","3"]}]"""),
            output.toString(),
            "A single document list was not represented correctly"
        )
    }

    fun writeJsonArrayMany(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        val output = StringWriter()
        val writer = PrintWriter(output)
        db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE) + orderBy(listOf(Field.named("id"))), listOf(), writer,
            Results::jsonFromData)
        assertEquals(
            JsonFunctions.maybeJsonB("""[{"id":"first","values":["a","b","c"]},"""
                    + """{"id":"second","values":["c","d","e"]},{"id":"third","values":["x","y","z"]}]"""),
            output.toString(),
            "A multiple document list was not represented correctly"
        )
    }

    fun singleNone(db: ThrowawayDatabase) =
        assertNull(
            db.conn.customSingle(FindQuery.all(TEST_TABLE), mapFunc = Results::fromData),
            "There should not have been a document returned"
        )

    fun singleOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertNotNull(
            db.conn.customSingle<JsonDocument>(FindQuery.all(TEST_TABLE), mapFunc = Results::fromData),
            "There should not have been a document returned"
        )
    }

    fun jsonSingleNone(db: ThrowawayDatabase) =
        assertEquals("{}", db.conn.customJsonSingle(FindQuery.all(TEST_TABLE), listOf(), Results::jsonFromData),
            "An empty document was not represented correctly")

    fun jsonSingleOne(db: ThrowawayDatabase) {
        db.conn.insert(TEST_TABLE, ArrayDocument("me", listOf("myself", "i")))
        assertEquals(
            JsonFunctions.maybeJsonB("{\"id\":\"me\",\"values\":[\"myself\",\"i\"]}"),
            db.conn.customJsonSingle(FindQuery.all(TEST_TABLE), listOf(), Results::jsonFromData),
            "A single document was not represented correctly"
        )
    }

    fun nonQueryChanges(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            5L, db.conn.customScalar(CountQuery.all(TEST_TABLE), mapFunc = Results::toCount),
            "There should have been 5 documents in the table"
        )
        db.conn.customNonQuery("DELETE FROM $TEST_TABLE")
        assertEquals(
            0L, db.conn.customScalar(CountQuery.all(TEST_TABLE), mapFunc = Results::toCount),
            "There should have been no documents in the table"
        )
    }

    fun nonQueryNoChanges(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            5L, db.conn.customScalar(CountQuery.all(TEST_TABLE), mapFunc = Results::toCount),
            "There should have been 5 documents in the table"
        )
        db.conn.customNonQuery(
            DeleteQuery.byId(TEST_TABLE, "eighty-two"),
            listOf(Parameter(":id", ParameterType.STRING, "eighty-two"))
        )
        assertEquals(
            5L, db.conn.customScalar(CountQuery.all(TEST_TABLE), mapFunc = Results::toCount),
            "There should still have been 5 documents in the table"
        )
    }

    fun scalar(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            3L,
            db.conn.customScalar("SELECT 3 AS it FROM $TEST_TABLE LIMIT 1", mapFunc = Results::toCount),
            "The number 3 should have been returned"
        )
    }

}