package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Field, FieldMatch}
import solutions.bitbadger.documents.java.Count as CoreCount

import java.sql.Connection

import _root_.scala.jdk.CollectionConverters.*

/**
 * Functions to count documents
 */
object Count:

  /**
   * Count all documents in the table
   *
   * @param tableName The name of the table in which documents should be counted
   * @param conn The connection over which documents should be counted
   * @return A count of the documents in the table
   * @throws DocumentException If any dependent process does
   */
  def all(tableName: String, conn: Connection): Long =
    CoreCount.all(tableName, conn)

  /**
   * Count all documents in the table (creates connection)
   *
   * @param tableName The name of the table in which documents should be counted
   * @return A count of the documents in the table
   * @throws DocumentException If no connection string has been set
   */
  def all(tableName: String): Long =
    CoreCount.all(tableName)

  /**
   * Count documents using a field comparison
   *
   * @param tableName  The name of the table in which documents should be counted
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, default `ALL`)
   * @param conn       The connection on which the deletion should be executed
   * @return A count of the matching documents in the table
   * @throws DocumentException If no dialect has been configured
   */
  def byFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch], conn: Connection): Long =
    CoreCount.byFields(tableName, fields.asJava, howMatched.orNull, conn)

  /**
   * Count documents using a field comparison (creates connection)
   *
   * @param tableName  The name of the table in which documents should be counted
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @return A count of the matching documents in the table
   * @throws DocumentException If no connection string has been set
   */
  def byFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Long =
    CoreCount.byFields(tableName, fields.asJava, howMatched.orNull)

  /**
   * Count documents using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be counted
   * @param criteria  The object for which JSON containment should be checked
   * @param conn      The connection on which the count should be executed
   * @return A count of the matching documents in the table
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A, conn: Connection): Long =
    CoreCount.byContains(tableName, criteria, conn)

  /**
   * Count documents using a JSON containment query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which documents should be counted
   * @param criteria  The object for which JSON containment should be checked
   * @return A count of the matching documents in the table
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A): Long =
    CoreCount.byContains(tableName, criteria)

  /**
   * Count documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which documents should be counted
   * @param path      The JSON path comparison to match
   * @param conn      The connection on which the count should be executed
   * @return A count of the matching documents in the table
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String, conn: Connection): Long =
    CoreCount.byJsonPath(tableName, path, conn)

  /**
   * Count documents using a JSON Path match query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which documents should be counted
   * @param path      The JSON path comparison to match
   * @return A count of the matching documents in the table
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String): Long =
    CoreCount.byJsonPath(tableName, path)
