package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Configuration, Field, FieldMatch, Parameter, ParameterType}
import solutions.bitbadger.documents.query.{FindQuery, QueryUtils}

import java.sql.Connection
import scala.reflect.ClassTag
import scala.jdk.CollectionConverters.*
import scala.util.Using

/**
 * Functions to find and retrieve documents
 */
object Find:

  /**
   * Retrieve all documents in the given table, ordering results by the optional given fields
   *
   * @param tableName The table from which documents should be retrieved
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents from the given table
   * @throws DocumentException If query execution fails
   */
  def all[Doc](tableName: String, orderBy: Seq[Field[?]], conn: Connection)(using tag: ClassTag[Doc]): List[Doc] =
    Custom.list[Doc](FindQuery.all(tableName) + QueryUtils.orderBy(orderBy.asJava), conn, Results.fromData)

  /**
   * Retrieve all documents in the given table, ordering results by the optional given fields
   *
   * @param tableName The table from which documents should be retrieved
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents from the given table
   * @throws DocumentException If query execution fails
   */
  def all[Doc](tableName: String, conn: Connection)(using tag: ClassTag[Doc]): List[Doc] =
    all[Doc](tableName, List(), conn)
    
  /**
   * Retrieve all documents in the given table (creates connection)
   *
   * @param tableName The table from which documents should be retrieved
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents from the given table
   * @throws DocumentException If no connection string has been set, or if query execution fails
   */
  def all[Doc](tableName: String, orderBy: Seq[Field[?]] = Nil)(using tag: ClassTag[Doc]): List[Doc] =
    Using(Configuration.dbConn()) { conn => all[Doc](tableName, orderBy, conn) }.get

  /**
   * Retrieve a document by its ID
   *
   * @param tableName The table from which the document should be retrieved
   * @param docId     The ID of the document to retrieve
   * @param conn      The connection over which documents should be retrieved
   * @return An `Option` with the document if it is found
   * @throws DocumentException If no dialect has been configured
   */
  def byId[Key, Doc](tableName: String, docId: Key, conn: Connection)(using tag: ClassTag[Doc]): Option[Doc] =
    Custom.single[Doc](FindQuery.byId(tableName, docId),
      Parameters.addFields(Field.equal(Configuration.idField, docId, ":id") :: Nil).toSeq, conn, Results.fromData)

  /**
   * Retrieve a document by its ID (creates connection)
   *
   * @param tableName The table from which the document should be retrieved
   * @param docId     The ID of the document to retrieve
   * @return An `Option` with the document if it is found
   * @throws DocumentException If no connection string has been set
   */
  def byId[Key, Doc](tableName: String, docId: Key)(using tag: ClassTag[Doc]): Option[Doc] =
    Using(Configuration.dbConn()) { conn => byId[Key, Doc](tableName, docId, conn) }.get

  /**
   * Retrieve documents using a field comparison, ordering results by the given fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn       The connection over which documents should be retrieved
   * @return A list of documents matching the field comparison
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch], orderBy: Seq[Field[?]],
                    conn: Connection)(using tag: ClassTag[Doc]): List[Doc] =
    val named = Parameters.nameFields(fields)
    Custom.list[Doc](
      FindQuery.byFields(tableName, named.asJava, howMatched.orNull) + QueryUtils.orderBy(orderBy.asJava),
      Parameters.addFields(named).toSeq, conn, Results.fromData)

  /**
   * Retrieve documents using a field comparison
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param conn       The connection over which documents should be retrieved
   * @return A list of documents matching the field comparison
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch], conn: Connection)
                   (using tag: ClassTag[Doc]): List[Doc] =
    byFields[Doc](tableName, fields, howMatched, List(), conn)

  /**
   * Retrieve documents using a field comparison, ordering results by the given fields
   *
   * @param tableName The table from which documents should be retrieved
   * @param fields    The fields which should be compared
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents matching the field comparison
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields[Doc](tableName: String, fields: Seq[Field[?]], orderBy: Seq[Field[?]], conn: Connection)
                   (using tag: ClassTag[Doc]): List[Doc] =
    byFields[Doc](tableName, fields, None, orderBy, conn)

  /**
   * Retrieve documents using a field comparison, ordering results by the given fields (creates connection)
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents matching the field comparison
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def byFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                    orderBy: Seq[Field[?]] = Nil)(using tag: ClassTag[Doc]): List[Doc] =
    Using(Configuration.dbConn()) { conn => byFields[Doc](tableName, fields, howMatched, orderBy, conn) }.get

  /**
   * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents matching the JSON containment query
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[Doc, A](tableName: String, criteria: A, orderBy: Seq[Field[?]], conn: Connection)
                        (using tag: ClassTag[Doc]): List[Doc] =
    Custom.list[Doc](FindQuery.byContains(tableName) + QueryUtils.orderBy(orderBy.asJava),
      Parameters.json(":criteria", criteria) :: Nil, conn, Results.fromData)

  /**
   * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents matching the JSON containment query
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[Doc, A](tableName: String, criteria: A, conn: Connection)(using tag: ClassTag[Doc]): List[Doc] =
    byContains[Doc, A](tableName, criteria, List(), conn)

  /**
   * Retrieve documents using a JSON containment query, ordering results by the given fields (PostgreSQL only; creates
   * connection)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents matching the JSON containment query
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byContains[Doc, A](tableName: String, criteria: A, orderBy: Seq[Field[?]] = Nil)
                        (using tag: ClassTag[Doc]): List[Doc] =
    Using(Configuration.dbConn()) { conn => byContains[Doc, A](tableName, criteria, orderBy, conn) }.get

  /**
   * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents matching the JSON Path match query
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath[Doc](tableName: String, path: String, orderBy: Seq[Field[?]], conn: Connection)
                     (using tag: ClassTag[Doc]): List[Doc] =
    Custom.list[Doc](FindQuery.byJsonPath(tableName) + QueryUtils.orderBy(orderBy.asJava),
      Parameter(":path", ParameterType.STRING, path) :: Nil, conn, Results.fromData)

  /**
   * Retrieve documents using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param conn      The connection over which documents should be retrieved
   * @return A list of documents matching the JSON Path match query
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath[Doc](tableName: String, path: String, conn: Connection)(using tag: ClassTag[Doc]): List[Doc] =
    byJsonPath[Doc](tableName, path, List(), conn)

  /**
   * Retrieve documents using a JSON Path match query, ordering results by the given fields (PostgreSQL only; creates
   * connection)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return A list of documents matching the JSON Path match query
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byJsonPath[Doc](tableName: String, path: String, orderBy: Seq[Field[?]] = Nil)
                     (using tag: ClassTag[Doc]): List[Doc] =
    Using(Configuration.dbConn()) { conn => byJsonPath[Doc](tableName, path, orderBy, conn) }.get

  /**
   * Retrieve the first document using a field comparison and ordering fields
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn       The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the field comparison if found
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def firstByFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch],
                         orderBy: Seq[Field[?]], conn: Connection)(using tag: ClassTag[Doc]): Option[Doc] =
    val named = Parameters.nameFields(fields)
    Custom.single[Doc](
      FindQuery.byFields(tableName, named.asJava, howMatched.orNull) + QueryUtils.orderBy(orderBy.asJava),
      Parameters.addFields(named).toSeq, conn, Results.fromData)

  /**
   * Retrieve the first document using a field comparison
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
   * @param conn       The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the field comparison if found
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def firstByFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch], conn: Connection)
                        (using tag: ClassTag[Doc]): Option[Doc] =
    firstByFields[Doc](tableName, fields, howMatched, List(), conn)

  /**
   * Retrieve the first document using a field comparison and ordering fields
   *
   * @param tableName The table from which documents should be retrieved
   * @param fields    The fields which should be compared
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the field comparison if found
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def firstByFields[Doc](tableName: String, fields: Seq[Field[?]], orderBy: Seq[Field[?]], conn: Connection)
                        (using tag: ClassTag[Doc]): Option[Doc] =
    firstByFields[Doc](tableName, fields, None, orderBy, conn)

  /**
   * Retrieve the first document using a field comparison
   *
   * @param tableName The table from which documents should be retrieved
   * @param fields    The fields which should be compared
   * @param conn      The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the field comparison if found
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def firstByFields[Doc](tableName: String, fields: Seq[Field[?]], conn: Connection)
                        (using tag: ClassTag[Doc]): Option[Doc] =
    firstByFields[Doc](tableName, fields, None, List(), conn)

  /**
   * Retrieve the first document using a field comparison and optional ordering fields (creates connection)
   *
   * @param tableName  The table from which documents should be retrieved
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
   * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return An `Option` with the first document matching the field comparison if found
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def firstByFields[Doc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                         orderBy: Seq[Field[?]] = Nil)(using tag: ClassTag[Doc]): Option[Doc] =
    Using(Configuration.dbConn()) { conn => firstByFields[Doc](tableName, fields, howMatched, orderBy, conn) }.get

  /**
   * Retrieve the first document using a JSON containment query and ordering fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the JSON containment query if found
   * @throws DocumentException If called on a SQLite connection
   */
  def firstByContains[Doc, A](tableName: String, criteria: A, orderBy: Seq[Field[?]], conn: Connection)
                             (using tag: ClassTag[Doc]): Option[Doc] =
    Custom.single[Doc](FindQuery.byContains(tableName) + QueryUtils.orderBy(orderBy.asJava),
      Parameters.json(":criteria", criteria) :: Nil, conn, Results.fromData)

  /**
   * Retrieve the first document using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param conn      The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the JSON containment query if found
   * @throws DocumentException If called on a SQLite connection
   */
  def firstByContains[Doc, A](tableName: String, criteria: A, conn: Connection)(using tag: ClassTag[Doc]): Option[Doc] =
    firstByContains[Doc, A](tableName, criteria, List(), conn)

  /**
   * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only; creates
   * connection)
   *
   * @param tableName The table from which documents should be retrieved
   * @param criteria  The object for which JSON containment should be checked
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return An `Option` with the first document matching the JSON containment query if found
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def firstByContains[Doc, A](tableName: String, criteria: A, orderBy: Seq[Field[?]] = Nil)
                             (using tag: ClassTag[Doc]): Option[Doc] =
    Using(Configuration.dbConn()) { conn => firstByContains[Doc, A](tableName, criteria, orderBy, conn) }.get

  /**
   * Retrieve the first document using a JSON Path match query and ordering fields (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @param conn      The connection over which documents should be retrieved
   * @return An `Optional` item, with the first document matching the JSON Path match query if found
   * @throws DocumentException If called on a SQLite connection
   */
  def firstByJsonPath[Doc](tableName: String, path: String, orderBy: Seq[Field[?]], conn: Connection)
                          (using tag: ClassTag[Doc]): Option[Doc] =
    Custom.single[Doc](FindQuery.byJsonPath(tableName) + QueryUtils.orderBy(orderBy.asJava),
      Parameter(":path", ParameterType.STRING, path) :: Nil, conn, Results.fromData)

  /**
   * Retrieve the first document using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param conn      The connection over which documents should be retrieved
   * @return An `Option` with the first document matching the JSON Path match query if found
   * @throws DocumentException If called on a SQLite connection
   */
  def firstByJsonPath[Doc](tableName: String, path: String, conn: Connection)(using tag: ClassTag[Doc]): Option[Doc] =
    firstByJsonPath[Doc](tableName, path, List(), conn)

  /**
   * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only; creates
   * connection)
   *
   * @param tableName The table from which documents should be retrieved
   * @param path      The JSON path comparison to match
   * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
   * @return An `Optional` item, with the first document matching the JSON Path match query if found
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def firstByJsonPath[Doc](tableName: String, path: String, orderBy: Seq[Field[?]] = Nil)
                          (using tag: ClassTag[Doc]): Option[Doc] =
    Using(Configuration.dbConn()) { conn => firstByJsonPath[Doc](tableName, path, orderBy, conn) }.get
