package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.java.Results as CoreResults

import java.io.PrintWriter
import java.sql.{PreparedStatement, ResultSet, SQLException}
import scala.collection.mutable.ListBuffer
import scala.reflect.ClassTag
import scala.util.Using

/**
 * Functions to manipulate results
 */
object Results:

  /**
   * Create a domain item from a document, specifying the field in which the document is found
   *
   * @param field   The field name containing the JSON document
   * @param rs      A `ResultSet` set to the row with the document to be constructed
   * @param ignored The class tag (placeholder used for signature; implicit tag used for serialization)
   * @return The constructed domain item
   */
  def fromDocument[Doc](field: String, rs: ResultSet, ignored: ClassTag[Doc])(using tag: ClassTag[Doc]): Doc =
    CoreResults.fromDocument(field, rs, tag.runtimeClass.asInstanceOf[Class[Doc]])

  /**
   * Create a domain item from a document
   *
   * @param rs      A `ResultSet` set to the row with the document to be constructed
   * @param ignored The class tag (placeholder used for signature; implicit tag used for serialization)
   * @return The constructed domain item
   */
  def fromData[Doc](rs: ResultSet, ignored: ClassTag[Doc])(using tag: ClassTag[Doc]): Doc =
    fromDocument[Doc]("data", rs, tag)

  /**
   * Create a list of items for the results of the given command, using the specified mapping function
   *
   * @param stmt    The prepared statement to execute
   * @param mapFunc The mapping function from data reader to domain class instance
   * @return A list of items from the query's result
   * @throws DocumentException If there is a problem executing the query (unchecked)
   */
  def toCustomList[Doc](stmt: PreparedStatement,mapFunc: (ResultSet, ClassTag[Doc]) => Doc)
                       (using tag: ClassTag[Doc]): List[Doc] =
    try
      val buffer = ListBuffer[Doc]()
      Using(stmt.executeQuery()) { rs =>
        while rs.next() do
          buffer.append(mapFunc(rs, tag))
      }
      buffer.toList
    catch
      case ex: SQLException =>
          throw DocumentException("Error retrieving documents from query: ${ex.message}", ex)

  /**
   * Extract a count from the first column
   *
   * @param rs A `ResultSet` set to the row with the count to retrieve
   * @return The count from the row
   * @throws DocumentException If the dialect has not been set (unchecked)
   */
  def toCount(rs: ResultSet, tag: ClassTag[Long] = ClassTag.Long): Long =
    CoreResults.toCount(rs, Long.getClass)

  /**
   * Extract a true/false value from the first column
   *
   * @param rs A `ResultSet` set to the row with the true/false value to retrieve
   * @return The true/false value from the row
   * @throws DocumentException If the dialect has not been set (unchecked)
   */
  def toExists(rs: ResultSet, tag: ClassTag[Boolean] = ClassTag.Boolean): Boolean =
    CoreResults.toExists(rs, Boolean.getClass)

  /**
   * Retrieve the JSON text of a document, specifying the field in which the document is found
   *
   * @param field The field name containing the JSON document
   * @param rs    A `ResultSet` set to the row with the document to be constructed
   * @return The JSON text of the document
   */
  def jsonFromDocument(field: String, rs: ResultSet): String =
    CoreResults.jsonFromDocument(field, rs)
  
  /**
   * Retrieve the JSON text of a document, specifying the field in which the document is found
   *
   * @param rs A `ResultSet` set to the row with the document to be constructed
   * @return The JSON text of the document
   */
  def jsonFromData(rs: ResultSet): String =
    CoreResults.jsonFromData(rs)
  
  /**
   * Create a JSON array of items for the results of the given command, using the specified mapping function
   *
   * @param stmt    The prepared statement to execute
   * @param mapFunc The mapping function from data reader to JSON text
   * @return A string with a JSON array of documents from the query's result
   * @throws DocumentException If there is a problem executing the query (unchecked)
   */
  def toJsonArray(stmt: PreparedStatement, mapFunc: ResultSet => String): String =
    try
      val results = StringBuilder("[")
      Using(stmt.executeQuery()) { rs =>
        while rs.next() do
          if (results.length > 2) results.append(",")
          results.append(mapFunc(rs))
      }
      results.append("]").toString()
    catch
      case ex: SQLException => throw DocumentException("Error retrieving documents from query: ${ex.message}", ex)

  /**
   * Write a JSON array of items for the results of the given command to the given `PrintWriter`, using the specified
   * mapping function
   *
   * @param writer  The writer for the results of the query
   * @param stmt    The prepared statement to execute
   * @param mapFunc The mapping function from data reader to JSON text
   * @return A string with a JSON array of documents from the query's result
   * @throws DocumentException If there is a problem executing the query (unchecked)
   */
  def writeJsonArray(writer: PrintWriter, stmt: PreparedStatement, mapFunc: ResultSet => String): Unit =
    try
      writer.write("[")
      Using(stmt.executeQuery()) { rs =>
        var isFirst = true
        while rs.next() do
          if isFirst then
            isFirst = false
          else
            writer.write(",")
          writer.write(mapFunc(rs))
      }
      writer.write("]")
    catch
      case ex: SQLException => throw DocumentException("Error writing documents from query: ${ex.message}", ex)

