package solutions.bitbadger.documents.scala.tests

import org.junit.jupiter.api.Assertions.*
import org.junit.jupiter.api.{AfterEach, DisplayName, Test}
import solutions.bitbadger.documents.*

import _root_.scala.jdk.CollectionConverters.*

@DisplayName("Scala | Field")
class FieldTest:

  /**
   * Clear the connection string (resets Dialect)
   */
  @AfterEach
  def cleanUp(): Unit =
    ForceDialect.none()

  // ~~~ INSTANCE METHODS ~~~

  @Test
  @DisplayName("withParameterName fails for invalid name")
  def withParamNameFails(): Unit =
    assertThrows(classOf[DocumentException], () => Field.equal("it", "").withParameterName("2424"))

  @Test
  @DisplayName("withParameterName works with colon prefix")
  def withParamNameColon(): Unit =
    val field = Field.equal("abc", "22").withQualifier("me")
    val withParam = field.withParameterName(":test")
    assertNotSame(field, withParam, "A new Field instance should have been created")
    assertEquals(field.getName, withParam.getName, "Name should have been preserved")
    assertEquals(field.getComparison, withParam.getComparison, "Comparison should have been preserved")
    assertEquals(":test", withParam.getParameterName, "Parameter name not set correctly")
    assertEquals(field.getQualifier, withParam.getQualifier, "Qualifier should have been preserved")

  @Test
  @DisplayName("withParameterName works with at-sign prefix")
  def withParamNameAtSign(): Unit =
    val field = Field.equal("def", "44")
    val withParam = field.withParameterName("@unit")
    assertNotSame(field, withParam, "A new Field instance should have been created")
    assertEquals(field.getName, withParam.getName, "Name should have been preserved")
    assertEquals(field.getComparison, withParam.getComparison, "Comparison should have been preserved")
    assertEquals("@unit", withParam.getParameterName, "Parameter name not set correctly")
    assertEquals(field.getQualifier, withParam.getQualifier, "Qualifier should have been preserved")

  @Test
  @DisplayName("withQualifier sets qualifier correctly")
  def withQualifier(): Unit =
    val field = Field.equal("j", "k")
    val withQual = field.withQualifier("test")
    assertNotSame(field, withQual, "A new Field instance should have been created")
    assertEquals(field.getName, withQual.getName, "Name should have been preserved")
    assertEquals(field.getComparison, withQual.getComparison, "Comparison should have been preserved")
    assertEquals(field.getParameterName, withQual.getParameterName, "Parameter Name should have been preserved")
    assertEquals("test", withQual.getQualifier, "Qualifier not set correctly")

  @Test
  @DisplayName("path generates for simple unqualified PostgreSQL field")
  def pathPostgresSimpleUnqualified(): Unit =
    assertEquals("data->>'SomethingCool'",
      Field.greaterOrEqual("SomethingCool", 18).path(Dialect.POSTGRESQL, FieldFormat.SQL), "Path not correct")

  @Test
  @DisplayName("path generates for simple qualified PostgreSQL field")
  def pathPostgresSimpleQualified(): Unit =
    assertEquals("this.data->>'SomethingElse'",
      Field.less("SomethingElse", 9).withQualifier("this").path(Dialect.POSTGRESQL, FieldFormat.SQL),
      "Path not correct")

  @Test
  @DisplayName("path generates for nested unqualified PostgreSQL field")
  def pathPostgresNestedUnqualified(): Unit =
    assertEquals("data#>>'{My,Nested,Field}'",
      Field.equal("My.Nested.Field", "howdy").path(Dialect.POSTGRESQL, FieldFormat.SQL), "Path not correct")

  @Test
  @DisplayName("path generates for nested qualified PostgreSQL field")
  def pathPostgresNestedQualified(): Unit =
    assertEquals("bird.data#>>'{Nest,Away}'",
      Field.equal("Nest.Away", "doc").withQualifier("bird").path(Dialect.POSTGRESQL, FieldFormat.SQL),
      "Path not correct")

  @Test
  @DisplayName("path generates for simple unqualified SQLite field")
  def pathSQLiteSimpleUnqualified(): Unit =
    assertEquals("data->>'SomethingCool'",
      Field.greaterOrEqual("SomethingCool", 18).path(Dialect.SQLITE, FieldFormat.SQL), "Path not correct")

  @Test
  @DisplayName("path generates for simple qualified SQLite field")
  def pathSQLiteSimpleQualified(): Unit =
    assertEquals("this.data->>'SomethingElse'",
      Field.less("SomethingElse", 9).withQualifier("this").path(Dialect.SQLITE, FieldFormat.SQL),
      "Path not correct")

  @Test
  @DisplayName("path generates for nested unqualified SQLite field")
  def pathSQLiteNestedUnqualified(): Unit =
    assertEquals("data->'My'->'Nested'->>'Field'",
      Field.equal("My.Nested.Field", "howdy").path(Dialect.SQLITE, FieldFormat.SQL), "Path not correct")

  @Test
  @DisplayName("path generates for nested qualified SQLite field")
  def pathSQLiteNestedQualified(): Unit =
    assertEquals("bird.data->'Nest'->>'Away'",
      Field.equal("Nest.Away", "doc").withQualifier("bird").path(Dialect.SQLITE, FieldFormat.SQL),
      "Path not correct")

  @Test
  @DisplayName("toWhere generates for exists w/o qualifier | PostgreSQL")
  def toWhereExistsNoQualPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'that_field' IS NOT NULL", Field.exists("that_field").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for exists w/o qualifier | SQLite")
  def toWhereExistsNoQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'that_field' IS NOT NULL", Field.exists("that_field").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for not-exists w/o qualifier | PostgreSQL")
  def toWhereNotExistsNoQualPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'a_field' IS NULL", Field.notExists("a_field").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for not-exists w/o qualifier | SQLite")
  def toWhereNotExistsNoQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'a_field' IS NULL", Field.notExists("a_field").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for BETWEEN w/o qualifier, numeric range | PostgreSQL")
  def toWhereBetweenNoQualNumericPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("(data->>'age')::numeric BETWEEN @agemin AND @agemax",
      Field.between("age", 13, 17, "@age").toWhere, "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for BETWEEN w/o qualifier, alphanumeric range | PostgreSQL")
  def toWhereBetweenNoQualAlphaPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'city' BETWEEN :citymin AND :citymax",
      Field.between("city", "Atlanta", "Chicago", ":city").toWhere, "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for BETWEEN w/o qualifier | SQLite")
  def toWhereBetweenNoQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'age' BETWEEN @agemin AND @agemax", Field.between("age", 13, 17, "@age").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for BETWEEN w/ qualifier, numeric range | PostgreSQL")
  def toWhereBetweenQualNumericPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("(test.data->>'age')::numeric BETWEEN @agemin AND @agemax",
      Field.between("age", 13, 17, "@age").withQualifier("test").toWhere, "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for BETWEEN w/ qualifier, alphanumeric range | PostgreSQL")
  def toWhereBetweenQualAlphaPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("unit.data->>'city' BETWEEN :citymin AND :citymax",
      Field.between("city", "Atlanta", "Chicago", ":city").withQualifier("unit").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for BETWEEN w/ qualifier | SQLite")
  def toWhereBetweenQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("my.data->>'age' BETWEEN @agemin AND @agemax",
      Field.between("age", 13, 17, "@age").withQualifier("my").toWhere, "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for IN/any, numeric values | PostgreSQL")
  def toWhereAnyNumericPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("(data->>'even')::numeric IN (:nbr_0, :nbr_1, :nbr_2)",
      Field.any("even", List(2, 4, 6).asJava, ":nbr").toWhere, "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for IN/any, alphanumeric values | PostgreSQL")
  def toWhereAnyAlphaPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'test' IN (:city_0, :city_1)",
      Field.any("test", List("Atlanta", "Chicago").asJava, ":city").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for IN/any | SQLite")
  def toWhereAnySQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'test' IN (:city_0, :city_1)",
      Field.any("test", List("Atlanta", "Chicago").asJava, ":city").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for inArray | PostgreSQL")
  def toWhereInArrayPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->'even' ??| ARRAY[:it_0, :it_1, :it_2, :it_3]",
      Field.inArray("even", "tbl", List(2, 4, 6, 8).asJava, ":it").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for inArray | SQLite")
  def toWhereInArraySQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("EXISTS (SELECT 1 FROM json_each(tbl.data, '$.test') WHERE value IN (:city_0, :city_1))",
      Field.inArray("test", "tbl", List("Atlanta", "Chicago").asJava, ":city").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for others w/o qualifier | PostgreSQL")
  def toWhereOtherNoQualPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'some_field' = :value", Field.equal("some_field", "", ":value").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates for others w/o qualifier | SQLite")
  def toWhereOtherNoQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'some_field' = :value", Field.equal("some_field", "", ":value").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates no-parameter w/ qualifier | PostgreSQL")
  def toWhereNoParamWithQualPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("test.data->>'no_field' IS NOT NULL", Field.exists("no_field").withQualifier("test").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates no-parameter w/ qualifier | SQLite")
  def toWhereNoParamWithQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("test.data->>'no_field' IS NOT NULL", Field.exists("no_field").withQualifier("test").toWhere,
      "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates parameter w/ qualifier | PostgreSQL")
  def toWhereParamWithQualPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("(q.data->>'le_field')::numeric <= :it",
      Field.lessOrEqual("le_field", 18, ":it").withQualifier("q").toWhere, "Field WHERE clause not generated correctly")

  @Test
  @DisplayName("toWhere generates parameter w/ qualifier | SQLite")
  def toWhereParamWithQualSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("q.data->>'le_field' <= :it",
      Field.lessOrEqual("le_field", 18, ":it").withQualifier("q").toWhere,
      "Field WHERE clause not generated correctly")

  // ~~~ STATIC CONSTRUCTOR TESTS ~~~

  @Test
  @DisplayName("equal constructs a field w/o parameter name")
  def equalCtor(): Unit =
    val field = Field.equal("Test", 14)
    assertEquals("Test", field.getName, "Field name not filled correctly")
    assertEquals(Op.EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(14, field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("equal constructs a field w/ parameter name")
  def equalParameterCtor(): Unit =
    val field = Field.equal("Test", 14, ":w")
    assertEquals("Test", field.getName, "Field name not filled correctly")
    assertEquals(Op.EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(14, field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":w", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("greater constructs a field w/o parameter name")
  def greaterCtor(): Unit =
    val field = Field.greater("Great", "night")
    assertEquals("Great", field.getName, "Field name not filled correctly")
    assertEquals(Op.GREATER, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("night", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("greater constructs a field w/ parameter name")
  def greaterParameterCtor(): Unit =
    val field = Field.greater("Great", "night", ":yeah")
    assertEquals("Great", field.getName, "Field name not filled correctly")
    assertEquals(Op.GREATER, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("night", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":yeah", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("greaterOrEqual constructs a field w/o parameter name")
  def greaterOrEqualCtor(): Unit =
    val field = Field.greaterOrEqual("Nice", 88L)
    assertEquals("Nice", field.getName, "Field name not filled correctly")
    assertEquals(Op.GREATER_OR_EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(88L, field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("greaterOrEqual constructs a field w/ parameter name")
  def greaterOrEqualParameterCtor(): Unit =
    val field = Field.greaterOrEqual("Nice", 88L, ":nice")
    assertEquals("Nice", field.getName, "Field name not filled correctly")
    assertEquals(Op.GREATER_OR_EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(88L, field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":nice", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("less constructs a field w/o parameter name")
  def lessCtor(): Unit =
    val field = Field.less("Lesser", "seven")
    assertEquals("Lesser", field.getName, "Field name not filled correctly")
    assertEquals(Op.LESS, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("seven", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("less constructs a field w/ parameter name")
  def lessParameterCtor(): Unit =
    val field = Field.less("Lesser", "seven", ":max")
    assertEquals("Lesser", field.getName, "Field name not filled correctly")
    assertEquals(Op.LESS, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("seven", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":max", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("lessOrEqual constructs a field w/o parameter name")
  def lessOrEqualCtor(): Unit =
    val field = Field.lessOrEqual("Nobody", "KNOWS")
    assertEquals("Nobody", field.getName, "Field name not filled correctly")
    assertEquals(Op.LESS_OR_EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("KNOWS", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("lessOrEqual constructs a field w/ parameter name")
  def lessOrEqualParameterCtor(): Unit =
    val field = Field.lessOrEqual("Nobody", "KNOWS", ":nope")
    assertEquals("Nobody", field.getName, "Field name not filled correctly")
    assertEquals(Op.LESS_OR_EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("KNOWS", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":nope", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("notEqual constructs a field w/o parameter name")
  def notEqualCtor(): Unit =
    val field = Field.notEqual("Park", "here")
    assertEquals("Park", field.getName, "Field name not filled correctly")
    assertEquals(Op.NOT_EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("here", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("notEqual constructs a field w/ parameter name")
  def notEqualParameterCtor(): Unit =
    val field = Field.notEqual("Park", "here", ":now")
    assertEquals("Park", field.getName, "Field name not filled correctly")
    assertEquals(Op.NOT_EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("here", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":now", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("between constructs a field w/o parameter name")
  def betweenCtor(): Unit =
    val field = Field.between("Age", 18, 49)
    assertEquals("Age", field.getName, "Field name not filled correctly")
    assertEquals(Op.BETWEEN, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(18, field.getComparison.getValue.getFirst, "Field comparison min value not filled correctly")
    assertEquals(49, field.getComparison.getValue.getSecond, "Field comparison max value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("between constructs a field w/ parameter name")
  def betweenParameterCtor(): Unit =
    val field = Field.between("Age", 18, 49, ":limit")
    assertEquals("Age", field.getName, "Field name not filled correctly")
    assertEquals(Op.BETWEEN, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(18, field.getComparison.getValue.getFirst, "Field comparison min value not filled correctly")
    assertEquals(49, field.getComparison.getValue.getSecond, "Field comparison max value not filled correctly")
    assertEquals(":limit", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("any constructs a field w/o parameter name")
  def anyCtor(): Unit =
    val field = Field.any("Here", List(8, 16, 32).asJava)
    assertEquals("Here", field.getName, "Field name not filled correctly")
    assertEquals(Op.IN, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(List(8, 16, 32).asJava, field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("any constructs a field w/ parameter name")
  def anyParameterCtor(): Unit =
    val field = Field.any("Here", List(8, 16, 32).asJava, ":list")
    assertEquals("Here", field.getName, "Field name not filled correctly")
    assertEquals(Op.IN, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals(List(8, 16, 32).asJava, field.getComparison.getValue, "Field comparison value not filled correctly")
    assertEquals(":list", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("inArray constructs a field w/o parameter name")
  def inArrayCtor(): Unit =
    val field = Field.inArray("ArrayField", "table", List("z").asJava)
    assertEquals("ArrayField", field.getName, "Field name not filled correctly")
    assertEquals(Op.IN_ARRAY, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("table", field.getComparison.getValue.getFirst, "Field comparison table not filled correctly")
    assertEquals(List("z").asJava, field.getComparison.getValue.getSecond,
      "Field comparison values not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("inArray constructs a field w/ parameter name")
  def inArrayParameterCtor(): Unit =
    val field = Field.inArray("ArrayField", "table", List("z").asJava, ":a")
    assertEquals("ArrayField", field.getName, "Field name not filled correctly")
    assertEquals(Op.IN_ARRAY, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("table", field.getComparison.getValue.getFirst, "Field comparison table not filled correctly")
    assertEquals(List("z").asJava, field.getComparison.getValue.getSecond,
      "Field comparison values not filled correctly")
    assertEquals(":a", field.getParameterName, "Field parameter name not filled correctly")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("exists constructs a field")
  def existsCtor(): Unit =
    val field = Field.exists("Groovy")
    assertEquals("Groovy", field.getName, "Field name not filled correctly")
    assertEquals(Op.EXISTS, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("notExists constructs a field")
  def notExistsCtor(): Unit =
    val field = Field.notExists("Groovy")
    assertEquals("Groovy", field.getName, "Field name not filled correctly")
    assertEquals(Op.NOT_EXISTS, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("named constructs a field")
  def namedCtor(): Unit =
    val field = Field.named("Tacos")
    assertEquals("Tacos", field.getName, "Field name not filled correctly")
    assertEquals(Op.EQUAL, field.getComparison.getOp, "Field comparison operation not filled correctly")
    assertEquals("", field.getComparison.getValue, "Field comparison value not filled correctly")
    assertNull(field.getParameterName, "The parameter name should have been null")
    assertNull(field.getQualifier, "The qualifier should have been null")

  @Test
  @DisplayName("static constructors fail for invalid parameter name")
  def staticCtorsFailOnParamName(): Unit =
    assertThrows(classOf[DocumentException], () => Field.equal("a", "b", "that ain't it, Jack..."))

  @Test
  @DisplayName("nameToPath creates a simple PostgreSQL SQL name")
  def nameToPathPostgresSimpleSQL(): Unit =
    assertEquals("data->>'Simple'", Field.nameToPath("Simple", Dialect.POSTGRESQL, FieldFormat.SQL),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a simple SQLite SQL name")
  def nameToPathSQLiteSimpleSQL(): Unit =
    assertEquals("data->>'Simple'", Field.nameToPath("Simple", Dialect.SQLITE, FieldFormat.SQL),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a nested PostgreSQL SQL name")
  def nameToPathPostgresNestedSQL(): Unit =
    assertEquals("data#>>'{A,Long,Path,to,the,Property}'",
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.POSTGRESQL, FieldFormat.SQL),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a nested SQLite SQL name")
  def nameToPathSQLiteNestedSQL(): Unit =
    assertEquals("data->'A'->'Long'->'Path'->'to'->'the'->>'Property'",
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.SQLITE, FieldFormat.SQL),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a simple PostgreSQL JSON name")
  def nameToPathPostgresSimpleJSON(): Unit =
    assertEquals("data->'Simple'", Field.nameToPath("Simple", Dialect.POSTGRESQL, FieldFormat.JSON),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a simple SQLite JSON name")
  def nameToPathSQLiteSimpleJSON(): Unit =
    assertEquals("data->'Simple'", Field.nameToPath("Simple", Dialect.SQLITE, FieldFormat.JSON),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a nested PostgreSQL JSON name")
  def nameToPathPostgresNestedJSON(): Unit =
    assertEquals("data#>'{A,Long,Path,to,the,Property}'",
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.POSTGRESQL, FieldFormat.JSON),
      "Path not constructed correctly")

  @Test
  @DisplayName("nameToPath creates a nested SQLite JSON name")
  def nameToPathSQLiteNestedJSON(): Unit =
    assertEquals("data->'A'->'Long'->'Path'->'to'->'the'->'Property'",
      Field.nameToPath("A.Long.Path.to.the.Property", Dialect.SQLITE, FieldFormat.JSON),
      "Path not constructed correctly")
