package solutions.bitbadger.documents.scala.tests

import org.junit.jupiter.api.Assertions.*
import org.junit.jupiter.api.{AfterEach, DisplayName, Test}
import solutions.bitbadger.documents.{DocumentException, Field, FieldMatch}
import solutions.bitbadger.documents.query.Where

import scala.jdk.CollectionConverters.*

@DisplayName("Scala | Query | Where")
class WhereTest:

  /**
   * Clear the connection string (resets Dialect)
   */
  @AfterEach
  def cleanUp (): Unit =
    ForceDialect.none()

  @Test
  @DisplayName("byFields is blank when given no fields")
  def byFieldsBlankIfEmpty(): Unit =
    assertEquals("", Where.byFields(List().asJava))

  @Test
  @DisplayName("byFields generates one numeric field | PostgreSQL")
  def byFieldsOneFieldPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("(data->>'it')::numeric = :that", Where.byFields(List(Field.equal("it", 9, ":that")).asJava))

  @Test
  @DisplayName("byFields generates one alphanumeric field | PostgreSQL")
  def byFieldsOneAlphaFieldPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'it' = :that", Where.byFields(List(Field.equal("it", "", ":that")).asJava))

  @Test
  @DisplayName("byFields generates one field | SQLite")
  def byFieldsOneFieldSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'it' = :that", Where.byFields(List(Field.equal("it", "", ":that")).asJava))

  @Test
  @DisplayName("byFields generates multiple fields w/ default match | PostgreSQL")
  def byFieldsMultipleDefaultPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'1' = :one AND (data->>'2')::numeric = :two AND data->>'3' = :three",
      Where.byFields(
        List(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")).asJava))

  @Test
  @DisplayName("byFields generates multiple fields w/ default match | SQLite")
  def byFieldsMultipleDefaultSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'1' = :one AND data->>'2' = :two AND data->>'3' = :three",
      Where.byFields(
        List(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")).asJava))

  @Test
  @DisplayName("byFields generates multiple fields w/ ANY match | PostgreSQL")
  def byFieldsMultipleAnyPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'1' = :one OR (data->>'2')::numeric = :two OR data->>'3' = :three",
      Where.byFields(
        List(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")).asJava,
        FieldMatch.ANY))

  @Test
  @DisplayName("byFields generates multiple fields w/ ANY match | SQLite")
  def byFieldsMultipleAnySQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'1' = :one OR data->>'2' = :two OR data->>'3' = :three",
      Where.byFields(
        List(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")).asJava,
        FieldMatch.ANY))

  @Test
  @DisplayName("byId generates defaults for alphanumeric key | PostgreSQL")
  def byIdDefaultAlphaPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'id' = :id", Where.byId())

  @Test
  @DisplayName("byId generates defaults for numeric key | PostgreSQL")
  def byIdDefaultNumericPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("(data->>'id')::numeric = :id", Where.byId(":id", 5))

  @Test
  @DisplayName("byId generates defaults | SQLite")
  def byIdDefaultSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'id' = :id", Where.byId())

  @Test
  @DisplayName("byId generates named ID | PostgreSQL")
  def byIdDefaultNamedPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data->>'id' = :key", Where.byId(":key"))

  @Test
  @DisplayName("byId generates named ID | SQLite")
  def byIdDefaultNamedSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals("data->>'id' = :key", Where.byId(":key"))

  @Test
  @DisplayName("jsonContains generates defaults | PostgreSQL")
  def jsonContainsDefaultPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data @> :criteria", Where.jsonContains())

  @Test
  @DisplayName("jsonContains generates named parameter | PostgreSQL")
  def jsonContainsNamedPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("data @> :it", Where.jsonContains(":it"))

  @Test
  @DisplayName("jsonContains fails | SQLite")
  def jsonContainsFailsSQLite(): Unit =
    ForceDialect.sqlite()
    assertThrows(classOf[DocumentException], () => Where.jsonContains())

  @Test
  @DisplayName("jsonPathMatches generates defaults | PostgreSQL")
  def jsonPathMatchDefaultPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("jsonb_path_exists(data, :path::jsonpath)", Where.jsonPathMatches())

  @Test
  @DisplayName("jsonPathMatches generates named parameter | PostgreSQL")
  def jsonPathMatchNamedPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals("jsonb_path_exists(data, :jp::jsonpath)", Where.jsonPathMatches(":jp"))

  @Test
  @DisplayName("jsonPathMatches fails | SQLite")
  def jsonPathFailsSQLite(): Unit =
    ForceDialect.sqlite()
    assertThrows(classOf[DocumentException], () => Where.jsonPathMatches())
