package solutions.bitbadger.documents.scala.tests.integration

import org.junit.jupiter.api.Assertions.*
import solutions.bitbadger.documents.query.{CountQuery, DeleteQuery, FindQuery, QueryUtils}
import solutions.bitbadger.documents.scala.Results
import solutions.bitbadger.documents.scala.extensions.*
import solutions.bitbadger.documents.scala.tests.TEST_TABLE
import solutions.bitbadger.documents.{Configuration, Field, Parameter, ParameterType}

import java.io.{PrintWriter, StringWriter}
import scala.jdk.CollectionConverters.*

object CustomFunctions:

  def listEmpty(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    db.conn.deleteByFields(TEST_TABLE, Field.exists(Configuration.idField) :: Nil)
    val result = db.conn.customList[JsonDocument](FindQuery.all(TEST_TABLE), Results.fromData)
    assertEquals(0, result.size, "There should have been no results")

  def listAll(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val result = db.conn.customList[JsonDocument](FindQuery.all(TEST_TABLE), Results.fromData)
    assertEquals(5, result.size, "There should have been 5 results")

  def jsonArrayEmpty(db: ThrowawayDatabase): Unit =
    assertEquals(0L, db.conn.countAll(TEST_TABLE), "The test table should be empty")
    assertEquals("[]", db.conn.customJsonArray(FindQuery.all(TEST_TABLE), Nil, Results.jsonFromData),
      "An empty list was not represented correctly")

  def jsonArraySingle(db: ThrowawayDatabase): Unit =
    db.conn.insert(TEST_TABLE, ArrayDocument("one", "2" :: "3" :: Nil))
    assertEquals(JsonFunctions.maybeJsonB("""[{"id":"one","values":["2","3"]}]"""),
      db.conn.customJsonArray(FindQuery.all(TEST_TABLE), Nil, Results.jsonFromData),
      "A single document list was not represented correctly")
  
  def jsonArrayMany(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    assertEquals(JsonFunctions.maybeJsonB("""[{"id":"first","values":["a","b","c"]},"""
        + """{"id":"second","values":["c","d","e"]},{"id":"third","values":["x","y","z"]}]"""),
      db.conn.customJsonArray(FindQuery.all(TEST_TABLE) + QueryUtils.orderBy((Field.named("id") :: Nil).asJava), Nil,
        Results.jsonFromData),
      "A multiple document list was not represented correctly")

  def writeJsonArrayEmpty(db: ThrowawayDatabase): Unit =
    assertEquals(0L, db.conn.countAll(TEST_TABLE), "The test table should be empty")
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE), Nil, writer, Results.jsonFromData)
    assertEquals("[]", output.toString, "An empty list was not represented correctly")

  def writeJsonArraySingle(db: ThrowawayDatabase): Unit =
    db.conn.insert(TEST_TABLE, ArrayDocument("one", "2" :: "3" :: Nil))
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE), Nil, writer, Results.jsonFromData)
    assertEquals(JsonFunctions.maybeJsonB("""[{"id":"one","values":["2","3"]}]"""), output.toString,
      "A single document list was not represented correctly")

  def writeJsonArrayMany(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeCustomJsonArray(FindQuery.all(TEST_TABLE) + QueryUtils.orderBy((Field.named("id") :: Nil).asJava), Nil,
      writer, Results.jsonFromData)
    assertEquals(JsonFunctions.maybeJsonB("""[{"id":"first","values":["a","b","c"]},"""
        + """{"id":"second","values":["c","d","e"]},{"id":"third","values":["x","y","z"]}]"""),
      output.toString, "A multiple document list was not represented correctly")

  def singleNone(db: ThrowawayDatabase): Unit =
    assertTrue(db.conn.customSingle[JsonDocument](FindQuery.all(TEST_TABLE), Results.fromData).isEmpty,
      "There should not have been a document returned")

  def singleOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertTrue(db.conn.customSingle[JsonDocument](FindQuery.all(TEST_TABLE), Results.fromData).isDefined,
      "There should have been a document returned")

  def jsonSingleNone(db: ThrowawayDatabase): Unit =
    assertEquals("{}", db.conn.customJsonSingle(FindQuery.all(TEST_TABLE), Nil, Results.jsonFromData),
      "An empty document was not represented correctly")

  def jsonSingleOne(db: ThrowawayDatabase): Unit =
    db.conn.insert(TEST_TABLE, ArrayDocument("me", "myself" :: "i" :: Nil))
    assertEquals(JsonFunctions.maybeJsonB("{\"id\":\"me\",\"values\":[\"myself\",\"i\"]}"),
      db.conn.customJsonSingle(FindQuery.all(TEST_TABLE), Nil, Results.jsonFromData),
      "A single document was not represented correctly")

  def nonQueryChanges(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(5L, db.conn.customScalar[Long](CountQuery.all(TEST_TABLE), Results.toCount),
      "There should have been 5 documents in the table")
    db.conn.customNonQuery(s"DELETE FROM $TEST_TABLE")
    assertEquals(0L, db.conn.customScalar[Long](CountQuery.all(TEST_TABLE), Results.toCount),
      "There should have been no documents in the table")

  def nonQueryNoChanges(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(5L, db.conn.customScalar[Long](CountQuery.all(TEST_TABLE), Results.toCount),
      "There should have been 5 documents in the table")
    db.conn.customNonQuery(DeleteQuery.byId(TEST_TABLE, "eighty-two"),
      Parameter(":id", ParameterType.STRING, "eighty-two") :: Nil)
    assertEquals(5L, db.conn.customScalar[Long](CountQuery.all(TEST_TABLE), Results.toCount),
      "There should still have been 5 documents in the table")

  def scalar(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(3L, db.conn.customScalar[Long](s"SELECT 3 AS it FROM $TEST_TABLE LIMIT 1", Results.toCount),
      "The number 3 should have been returned")
