package solutions.bitbadger.documents.scala.tests.integration

import org.junit.jupiter.api.Assertions.*
import solutions.bitbadger.documents.{AutoId, Configuration, DocumentException, Field}
import solutions.bitbadger.documents.scala.extensions.*
import solutions.bitbadger.documents.scala.tests.TEST_TABLE

object DocumentFunctions:

  import org.junit.jupiter.api.Assertions.assertThrows

  def insertDefault(db: ThrowawayDatabase): Unit =
    assertEquals(0L, db.conn.countAll(TEST_TABLE), "There should be no documents in the table")
    val doc = JsonDocument("turkey", "", 0, SubDocument("gobble", "gobble"))
    db.conn.insert(TEST_TABLE, doc)
    val after = db.conn.findAll[JsonDocument](TEST_TABLE)
    assertEquals(1, after.size, "There should be one document in the table")
    assertEquals(doc, after.head, "The document should be what was inserted")

  def insertDupe(db: ThrowawayDatabase): Unit =
    db.conn.insert(TEST_TABLE, JsonDocument("a", "", 0, null))
    assertThrows(classOf[DocumentException], () => db.conn.insert(TEST_TABLE, JsonDocument("a", "b", 22, null)),
      "Inserting a document with a duplicate key should have thrown an exception")

  def insertNumAutoId(db: ThrowawayDatabase): Unit =
    try
      Configuration.autoIdStrategy = AutoId.NUMBER
      Configuration.idField = "key"
      assertEquals(0L, db.conn.countAll(TEST_TABLE), "There should be no documents in the table")

      db.conn.insert(TEST_TABLE, NumIdDocument(0, "one"))
      db.conn.insert(TEST_TABLE, NumIdDocument(0, "two"))
      db.conn.insert(TEST_TABLE, NumIdDocument(77, "three"))
      db.conn.insert(TEST_TABLE, NumIdDocument(0, "four"))

      val after = db.conn.findAll[NumIdDocument](TEST_TABLE, Field.named("key") :: Nil)
      assertEquals(4, after.size, "There should have been 4 documents returned")
      assertEquals("1|2|77|78", after.fold("") { (acc, item) => s"$acc|$item" }, "The IDs were not generated correctly")
    finally
      Configuration.autoIdStrategy = AutoId.DISABLED
      Configuration.idField = "id"

  def insertUUIDAutoId(db: ThrowawayDatabase): Unit =
    try
      Configuration.autoIdStrategy = AutoId.UUID
      assertEquals(0L, db.conn.countAll(TEST_TABLE), "There should be no documents in the table")

      db.conn.insert(TEST_TABLE, JsonDocument(""))

      val after = db.conn.findAll[JsonDocument](TEST_TABLE)
      assertEquals(1, after.size, "There should have been 1 document returned")
      assertEquals(32, after.head.id.length, "The ID was not generated correctly")
    finally
      Configuration.autoIdStrategy = AutoId.DISABLED

  def insertStringAutoId(db: ThrowawayDatabase): Unit =
    try
      Configuration.autoIdStrategy = AutoId.RANDOM_STRING
      assertEquals(0L, db.conn.countAll(TEST_TABLE), "There should be no documents in the table")

      db.conn.insert(TEST_TABLE, JsonDocument(""))

      Configuration.idStringLength = 21
      db.conn.insert(TEST_TABLE, JsonDocument(""))

      val after = db.conn.findAll[JsonDocument](TEST_TABLE)
      assertEquals(2, after.size, "There should have been 2 documents returned")
      assertEquals(16, after.head.id.length, "The first document's ID was not generated correctly")
      assertEquals(21, after(1).id.length, "The second document's ID was not generated correctly")
    finally
      Configuration.autoIdStrategy = AutoId.DISABLED
      Configuration.idStringLength = 16

  def saveMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    db.conn.save(TEST_TABLE, JsonDocument("two", numValue = 44))
    val tryDoc = db.conn.findById[String, JsonDocument](TEST_TABLE, "two")
    assertTrue(tryDoc.isDefined, "There should have been a document returned")
    val doc = tryDoc.get
    assertEquals("two", doc.id, "An incorrect document was returned")
    assertEquals("", doc.value, "The \"value\" field was not updated")
    assertEquals(44, doc.numValue, "The \"numValue\" field was not updated")
    assertNull(doc.sub, "The \"sub\" field was not updated")

  def saveNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    db.conn.save(TEST_TABLE, JsonDocument("test", sub = SubDocument("a", "b")))
    assertTrue(db.conn.findById[String, JsonDocument](TEST_TABLE, "test").isDefined,
      "The test document should have been saved")

  def updateMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    db.conn.update(TEST_TABLE, "one", JsonDocument("one", "howdy", 8, SubDocument("y", "z")))
    val tryDoc = db.conn.findById[String, JsonDocument](TEST_TABLE, "one")
    assertTrue(tryDoc.isDefined, "There should have been a document returned")
    val doc = tryDoc.get
    assertEquals("one", doc.id, "An incorrect document was returned")
    assertEquals("howdy", doc.value, "The \"value\" field was not updated")
    assertEquals(8, doc.numValue, "The \"numValue\" field was not updated")
    assertNotNull(doc.sub, "The sub-document should not be null")
    assertEquals("y", doc.sub.foo, "The sub-document \"foo\" field was not updated")
    assertEquals("z", doc.sub.bar, "The sub-document \"bar\" field was not updated")

  def updateNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.existsById(TEST_TABLE, "two-hundred"))
    db.conn.update(TEST_TABLE, "two-hundred", JsonDocument("two-hundred", numValue = 200))
    assertFalse(db.conn.existsById(TEST_TABLE, "two-hundred"))
