package solutions.bitbadger.documents.scala.tests.integration

import org.junit.jupiter.api.Assertions.*
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.scala.extensions.*
import solutions.bitbadger.documents.scala.tests.TEST_TABLE

object RemoveFieldsFunctions:

  def byIdMatchFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    db.conn.removeFieldsById(TEST_TABLE, "two", "sub" :: "value" :: Nil)
    val doc = db.conn.findById[String, JsonDocument](TEST_TABLE, "two")
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("", doc.get.value, "The value should have been empty")
    assertNull(doc.get.sub, "The sub-document should have been removed")

  def byIdMatchNoFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.existsByFields(TEST_TABLE, Field.exists("a_field_that_does_not_exist") :: Nil))
    db.conn.removeFieldsById(TEST_TABLE, "one", "a_field_that_does_not_exist" :: Nil) // no exception = pass

  def byIdNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.existsById(TEST_TABLE, "fifty"))
    db.conn.removeFieldsById(TEST_TABLE, "fifty", "sub" :: Nil) // no exception = pass

  def byFieldsMatchFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val fields = Field.equal("numValue", 17) :: Nil
    db.conn.removeFieldsByFields(TEST_TABLE, fields, "sub" :: Nil)
    val doc = db.conn.findFirstByFields[JsonDocument](TEST_TABLE, fields)
    assertTrue(doc.isDefined, "The document should have been returned")
    assertEquals("four", doc.get.id, "An incorrect document was returned")
    assertNull(doc.get.sub, "The sub-document should have been removed")

  def byFieldsMatchNoFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.existsByFields(TEST_TABLE, Field.exists("nada") :: Nil))
    db.conn.removeFieldsByFields(TEST_TABLE, Field.equal("numValue", 17) :: Nil, "nada" :: Nil) // no exn = pass

  def byFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val fields = Field.notEqual("missing", "nope") :: Nil
    assertFalse(db.conn.existsByFields(TEST_TABLE, fields))
    db.conn.removeFieldsByFields(TEST_TABLE, fields, "value" :: Nil) // no exception = pass

  def byContainsMatchFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val criteria = Map.Map1("sub", Map.Map1("foo", "green"))
    db.conn.removeFieldsByContains(TEST_TABLE, criteria, "value" :: Nil)
    val docs = db.conn.findByContains[JsonDocument, Map.Map1[String, Map.Map1[String, String]]](TEST_TABLE, criteria)
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    docs.foreach { doc =>
      assertTrue(("two" :: "four" :: Nil).contains(doc.id), s"An incorrect document was returned (${doc.id})")
      assertEquals("", doc.value, "The value should have been empty")
    }

  def byContainsMatchNoFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.existsByFields(TEST_TABLE, Field.exists("invalid_field") :: Nil))
    db.conn.removeFieldsByContains(TEST_TABLE, Map.Map1("sub", Map.Map1("foo", "green")), "invalid_field" :: Nil)
    // no exception = pass

  def byContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val contains = Map.Map1("value", "substantial")
    assertFalse(db.conn.existsByContains(TEST_TABLE, contains))
    db.conn.removeFieldsByContains(TEST_TABLE, contains, "numValue" :: Nil)

  def byJsonPathMatchFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val path = "$.value ? (@ == \"purple\")"
    db.conn.removeFieldsByJsonPath(TEST_TABLE, path, "sub" :: Nil)
    val docs = db.conn.findByJsonPath[JsonDocument](TEST_TABLE, path)
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    docs.foreach { doc =>
      assertTrue(("four" :: "five" :: Nil).contains(doc.id), s"An incorrect document was returned (${doc.id})")
      assertNull(doc.sub, "The sub-document should have been removed")
    }

  def byJsonPathMatchNoFields(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.existsByFields(TEST_TABLE, Field.exists("submarine") :: Nil))
    db.conn.removeFieldsByJsonPath(TEST_TABLE, "$.value ? (@ == \"purple\")", "submarine" :: Nil) // no exn = pass

  def byJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val path = "$.value ? (@ == \"mauve\")"
    assertFalse(db.conn.existsByJsonPath(TEST_TABLE, path))
    db.conn.removeFieldsByJsonPath(TEST_TABLE, path, "value" :: Nil) // no exception = pass
  